"""
Standard Responses
------------------
This module provides basic responses.
"""

import random
from typing import List

from chatsky.core import BaseResponse, Message, Context
from chatsky.core.message import MessageInitTypes


class RandomChoice(BaseResponse):
    """
    Return a random message from :py:attr:`responses`.
    """

    responses: List[Message]
    """A list of messages to choose from."""

    def __init__(self, *responses: MessageInitTypes):
        super().__init__(responses=responses)

    async def call(self, ctx: Context) -> MessageInitTypes:
        return random.choice(self.responses)



# -*- coding: utf-8 -*-
# flake8: noqa: F401
from importlib.metadata import version


__version__ = version(__name__)


import nest_asyncio as __nest_asyncio__

__nest_asyncio__.apply()

from chatsky.core import (
    GLOBAL,
    LOCAL,
    RESPONSE,
    TRANSITIONS,
    MISC,
    PRE_RESPONSE,
    PRE_TRANSITION,
    BaseCondition,
    AnyCondition,
    BaseResponse,
    AnyResponse,
    BaseDestination,
    AnyDestination,
    BaseProcessing,
    BasePriority,
    AnyPriority,
    Pipeline,
    Context,
    Message,
    Transition,
    Transition as Tr,
    MessageInitTypes,
    NodeLabel,
    NodeLabelInitTypes,
    AbsoluteNodeLabel,
    AbsoluteNodeLabelInitTypes,
)
import chatsky.conditions as cnd
import chatsky.destinations as dst
import chatsky.responses as rsp
import chatsky.processing as proc


import chatsky.__rebuild_pydantic_models__



"""
Actor
-----
Actor is a component of :py:class:`.Pipeline`, that processes the :py:class:`.Script`.

It is responsible for determining the next node and getting response from it.

The actor acts as a bridge between the user's input and the dialog graph,
making sure that the conversation follows the expected flow.

More details on the processing can be found in the documentation for
:py:meth:`Actor.run_component`.
"""

from __future__ import annotations
import logging
import asyncio
from typing import Dict

from chatsky.core.service.component import PipelineComponent
from chatsky.core.transition import get_next_label
from chatsky.core.message import Message

from chatsky.core.context import Context
from chatsky.core.script_function import BaseProcessing

logger = logging.getLogger(__name__)


class Actor(PipelineComponent):
    """
    The class which is used to process :py:class:`~chatsky.core.context.Context`
    according to the :py:class:`~chatsky.core.script.Script`.
    """

    @property
    def computed_name(self) -> str:
        """
        "actor"
        """
        return "actor"

    async def run_component(self, ctx: Context) -> None:
        """
        Process the context in the following way:

        1. Run pre-transition of the :py:attr:`.Context.current_node`.
        2. Determine and save the next node based on :py:attr:`~chatsky.core.script.Node.transitions`
           of the :py:attr:`.Context.current_node`.
        3. Run pre-response of the :py:attr:`.Context.current_node`.
        4. Determine and save the response of the :py:attr:`.Context.current_node`
        """
        next_label = ctx.pipeline.fallback_label

        try:
            ctx.framework_data.current_node = ctx.pipeline.script.get_inherited_node(ctx.last_label)

            logger.debug("Running pre_transition")
            await self._run_processing(ctx.current_node.pre_transition, ctx)

            logger.debug("Running transitions")

            destination_result = await get_next_label(ctx, ctx.current_node.transitions, ctx.pipeline.default_priority)
            if destination_result is not None:
                next_label = destination_result
        except Exception as exc:
            logger.exception("Exception occurred during transition processing.", exc_info=exc)

        logger.debug(f"Next label: {next_label}")

        ctx.add_label(next_label)

        response = Message()

        try:
            ctx.framework_data.current_node = ctx.pipeline.script.get_inherited_node(next_label)

            logger.debug("Running pre_response")
            await self._run_processing(ctx.current_node.pre_response, ctx)

            node_response = ctx.current_node.response
            if node_response is not None:
                response_result = await node_response.wrapped_call(ctx)
                if isinstance(response_result, Message):
                    response = response_result
                    logger.debug(f"Produced response {response}.")
                else:
                    logger.debug("Response was not produced.")
            else:
                logger.debug("Node has empty response.")
        except Exception as exc:
            logger.exception("Exception occurred during response processing.", exc_info=exc)

        ctx.add_response(response)

    @staticmethod
    async def _run_processing_parallel(processing: Dict[str, BaseProcessing], ctx: Context) -> None:
        """
        Execute :py:class:`.BaseProcessing` functions simultaneously, independent of the order.

        Picked depending on the value of the :py:class:`.Pipeline`'s `parallelize_processing` flag.
        """
        await asyncio.gather(
            *[func.wrapped_call(ctx, info=f"processing_name={name!r}") for name, func in processing.items()]
        )

    @staticmethod
    async def _run_processing_sequential(processing: Dict[str, BaseProcessing], ctx: Context) -> None:
        """
        Execute :py:class:`.BaseProcessing` functions in-order.

        Picked depending on the value of the :py:class:`.Pipeline`'s `parallelize_processing` flag.
        """
        for name, func in processing.items():
            await func.wrapped_call(ctx, info=f"processing_name={name!r}")

    @staticmethod
    async def _run_processing(processing: Dict[str, BaseProcessing], ctx: Context) -> None:
        """
        Run :py:class:`.BaseProcessing` functions.

        The execution order depends on the value of the :py:class:`.Pipeline`'s
        `parallelize_processing` flag.
        """
        if ctx.pipeline.parallelize_processing:
            await Actor._run_processing_parallel(processing, ctx)
        else:
            await Actor._run_processing_sequential(processing, ctx)



"""
Standard Processing
-------------------
This module provides basic processing functions.

- :py:class:`ModifyResponse` modifies response of the :py:attr:`.Context.current_node`.
"""

import abc

from chatsky.core import BaseProcessing, BaseResponse, Context, MessageInitTypes


class ModifyResponse(BaseProcessing, abc.ABC):
    """
    Modify the response function of the :py:attr:`.Context.current_node` to call
    :py:meth:`modified_response` instead.
    """

    @abc.abstractmethod
    async def modified_response(self, original_response: BaseResponse, ctx: Context) -> MessageInitTypes:
        """
        A function that replaces response of the current node.

        :param original_response: Response of the current node when :py:class:`.ModifyResponse` is called.
        :param ctx: Current context.
        """
        raise NotImplementedError

    async def call(self, ctx: Context) -> None:
        current_response = ctx.current_node.response
        if current_response is None:
            return

        processing_object = self

        class ModifiedResponse(BaseResponse):
            async def call(self, ctx: Context) -> MessageInitTypes:
                return await processing_object.modified_response(current_response, ctx)

        ctx.current_node.response = ModifiedResponse()



"""
Pipeline
--------
Pipeline is the main element of the Chatsky framework.

Pipeline is responsible for managing and executing the various components
(:py:class:`~chatsky.core.service.component.PipelineComponent`)
including :py:class:`.Actor`.
"""

import asyncio
import logging
from functools import cached_property
from typing import Union, List, Dict, Optional, Hashable
from pydantic import BaseModel, Field, model_validator, computed_field

from chatsky.context_storages import DBContextStorage
from chatsky.core.script import Script
from chatsky.core.context import Context
from chatsky.core.message import Message

from chatsky.messengers.console import CLIMessengerInterface
from chatsky.messengers.common import MessengerInterface
from chatsky.slots.slots import GroupSlot
from chatsky.core.service.group import ServiceGroup, ServiceGroupInitTypes
from chatsky.core.service.extra import ComponentExtraHandlerInitTypes, BeforeHandler, AfterHandler
from .service import Service
from .utils import finalize_service_group, initialize_service_states
from chatsky.core.service.actor import Actor
from chatsky.core.node_label import AbsoluteNodeLabel, AbsoluteNodeLabelInitTypes
from chatsky.core.script_parsing import JSONImporter, Path

logger = logging.getLogger(__name__)


class PipelineServiceGroup(ServiceGroup):
    """A service group that allows actor inside."""

    components: List[Union[Actor, Service, ServiceGroup]]


class Pipeline(BaseModel, extra="forbid", arbitrary_types_allowed=True):
    """
    Class that automates service execution and creates service pipeline.
    """

    pre_services: ServiceGroup = Field(default_factory=list, validate_default=True)
    """
    :py:class:`~.ServiceGroup` that will be executed before Actor.
    """
    post_services: ServiceGroup = Field(default_factory=list, validate_default=True)
    """
    :py:class:`~.ServiceGroup` that will be executed after :py:class:`~.Actor`.
    """
    script: Script
    """
    (required) A :py:class:`~.Script` instance (object or dict).
    """
    start_label: AbsoluteNodeLabel
    """
    (required) The first node of every context.
    """
    fallback_label: AbsoluteNodeLabel
    """
    Node which will is used if :py:class:`Actor` cannot find the next node.

    This most commonly happens when there are not suitable transitions.

    Defaults to :py:attr:`start_label`.
    """
    default_priority: float = 1.0
    """
    Default priority value for :py:class:`~chatsky.core.transition.Transition`.

    Defaults to ``1.0``.
    """
    slots: GroupSlot = Field(default_factory=GroupSlot)
    """
    Slots configuration.
    """
    messenger_interface: MessengerInterface = Field(default_factory=CLIMessengerInterface)
    """
    A `MessengerInterface` instance for this pipeline.

    It handles connections to interfaces that provide user requests and accept bot responses.
    """
    context_storage: Union[DBContextStorage, Dict] = Field(default_factory=dict)
    """
    A :py:class:`~.DBContextStorage` instance for this pipeline or
    a dict to store dialog :py:class:`~.Context`.
    """
    before_handler: BeforeHandler = Field(default_factory=list, validate_default=True)
    """
    :py:class:`~.BeforeHandler` to add to the pipeline service.
    """
    after_handler: AfterHandler = Field(default_factory=list, validate_default=True)
    """
    :py:class:`~.AfterHandler` to add to the pipeline service.
    """
    timeout: Optional[float] = None
    """
    Timeout to add to pipeline root service group.
    """
    parallelize_processing: bool = False
    """
    This flag determines whether or not the functions
    defined in the ``PRE_RESPONSE_PROCESSING`` and ``PRE_TRANSITIONS_PROCESSING`` sections
    of the script should be parallelized over respective groups.
    """

    def __init__(
        self,
        script: Union[Script, dict],
        start_label: AbsoluteNodeLabelInitTypes,
        fallback_label: AbsoluteNodeLabelInitTypes = None,
        *,
        default_priority: float = None,
        slots: GroupSlot = None,
        messenger_interface: MessengerInterface = None,
        context_storage: Union[DBContextStorage, dict] = None,
        pre_services: ServiceGroupInitTypes = None,
        post_services: ServiceGroupInitTypes = None,
        before_handler: ComponentExtraHandlerInitTypes = None,
        after_handler: ComponentExtraHandlerInitTypes = None,
        timeout: float = None,
        parallelize_processing: bool = None,
    ):
        if fallback_label is None:
            fallback_label = start_label
        init_dict = {
            "script": script,
            "start_label": start_label,
            "fallback_label": fallback_label,
            "default_priority": default_priority,
            "slots": slots,
            "messenger_interface": messenger_interface,
            "context_storage": context_storage,
            "pre_services": pre_services,
            "post_services": post_services,
            "before_handler": before_handler,
            "after_handler": after_handler,
            "timeout": timeout,
            "parallelize_processing": parallelize_processing,
        }
        empty_fields = set()
        for k, v in init_dict.items():
            if k not in self.model_fields:
                raise NotImplementedError("Init method contains a field not in model fields.")
            if v is None:
                empty_fields.add(k)
        for field in empty_fields:
            del init_dict[field]
        super().__init__(**init_dict)
        self.services_pipeline  # cache services

    @classmethod
    def from_file(
        cls,
        file: Union[str, Path],
        custom_dir: Union[str, Path] = "custom",
        **overrides,
    ) -> "Pipeline":
        """
        Create Pipeline by importing it from a file.
        A file (json or yaml) should contain a dictionary with keys being a subset of pipeline init parameters.

        See :py:meth:`.JSONImporter.import_pipeline_file` for more information.

        :param file: Path to a file containing pipeline init parameters.
        :param custom_dir: Path to a directory containing custom code.
            Defaults to "./custom".
            If ``file`` does not use custom code, this parameter will not have any effect.
        :param overrides: You can pass init parameters to override those imported from the ``file``.
        """
        pipeline = JSONImporter(custom_dir=custom_dir).import_pipeline_file(file)

        pipeline.update(overrides)

        return cls(**pipeline)

    @computed_field
    @cached_property
    def actor(self) -> Actor:
        """An actor instance of the pipeline."""
        return Actor()

    @computed_field
    @cached_property
    def services_pipeline(self) -> PipelineServiceGroup:
        """
        A group containing :py:attr:`.Pipeline.pre_services`, :py:class:`~.Actor`
        and :py:attr:`.Pipeline.post_services`.
        It has :py:attr:`.Pipeline.before_handler` and :py:attr:`.Pipeline.after_handler` applied to it.
        """
        components = [self.pre_services, self.actor, self.post_services]
        self.pre_services.name = "pre"
        self.post_services.name = "post"
        services_pipeline = PipelineServiceGroup(
            components=components,
            before_handler=self.before_handler,
            after_handler=self.after_handler,
            timeout=self.timeout,
        )
        services_pipeline.name = ""
        services_pipeline.path = ""

        finalize_service_group(services_pipeline, path=services_pipeline.path)

        return services_pipeline

    @model_validator(mode="after")
    def validate_start_label(self):
        """Validate :py:attr:`start_label` is in :py:attr:`script`."""
        if self.script.get_node(self.start_label) is None:
            raise ValueError(f"Unknown start_label={self.start_label}")
        return self

    @model_validator(mode="after")
    def validate_fallback_label(self):
        """Validate :py:attr:`fallback_label` is in :py:attr:`script`."""
        if self.script.get_node(self.fallback_label) is None:
            raise ValueError(f"Unknown fallback_label={self.fallback_label}")
        return self

    async def _run_pipeline(
        self, request: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that should be invoked on user input.
        This method has the same signature as :py:class:`~chatsky.core.service.types.PipelineRunnerFunction`.

        This method does:

        1. Retrieve from :py:attr:`context_storage` or initialize context ``ctx_id``.
        2. Update :py:attr:`.Context.misc` with ``update_ctx_misc``.
        3. Set up :py:attr:`.Context.framework_data` fields.
        4. Add ``request`` to the context.
        5. Execute :py:attr:`services_pipeline`.
           This includes :py:class:`.Actor` (read :py:meth:`.Actor.run_component` for more information).
        6. Save context in the :py:attr:`context_storage`.

        :return: Modified context ``ctx_id``.
        """
        logger.info(f"Running pipeline for context {ctx_id}.")
        logger.debug(f"Received request: {request}.")
        if ctx_id is None:
            ctx = Context.init(self.start_label)
        elif isinstance(self.context_storage, DBContextStorage):
            ctx = await self.context_storage.get_async(ctx_id, Context.init(self.start_label, id=ctx_id))
        else:
            ctx = self.context_storage.get(ctx_id, Context.init(self.start_label, id=ctx_id))

        if update_ctx_misc is not None:
            ctx.misc.update(update_ctx_misc)

        if self.slots is not None:
            ctx.framework_data.slot_manager.set_root_slot(self.slots)

        ctx.framework_data.pipeline = self
        initialize_service_states(ctx, self.services_pipeline)

        ctx.add_request(request)
        await self.services_pipeline(ctx)

        ctx.framework_data.service_states.clear()
        ctx.framework_data.pipeline = None

        if isinstance(self.context_storage, DBContextStorage):
            await self.context_storage.set_item_async(ctx_id, ctx)
        else:
            self.context_storage[ctx_id] = ctx

        return ctx

    def run(self):
        """
        Method that starts a pipeline and connects to :py:attr:`messenger_interface`.

        It passes :py:meth:`_run_pipeline` to :py:attr:`messenger_interface` as a callback,
        so every time user request is received, :py:meth:`_run_pipeline` will be called.

        This method can be both blocking and non-blocking. It depends on current :py:attr:`messenger_interface` nature.
        Message interfaces that run in a loop block current thread.
        """
        logger.info("Pipeline is accepting requests.")
        asyncio.run(self.messenger_interface.connect(self._run_pipeline))

    def __call__(
        self, request: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that executes pipeline once.
        Basically, it is a shortcut for :py:meth:`_run_pipeline`.
        NB! When pipeline is executed this way, :py:attr:`messenger_interface` won't be initiated nor connected.

        This method has the same signature as :py:class:`~chatsky.core.service.types.PipelineRunnerFunction`.
        """
        return asyncio.run(self._run_pipeline(request, ctx_id, update_ctx_misc))



"""
Service
-------
This module defines services -- a way to process context outside the Script.
"""

from .component import PipelineComponent
from .extra import BeforeHandler, AfterHandler
from .group import ServiceGroup
from .service import Service, to_service
from .types import (
    ExtraHandlerRuntimeInfo,
    ExtraHandlerType,
    PipelineRunnerFunction,
    ComponentExecutionState,
    ExtraHandlerConditionFunction,
    ExtraHandlerFunction,
)



"""
Component
---------
The Component module defines a :py:class:`.PipelineComponent` class.

This is a base class for pipeline processing and is responsible for performing a specific task.
"""

from __future__ import annotations

import logging
import abc
import asyncio
from typing import Optional, TYPE_CHECKING
from pydantic import BaseModel, Field, field_validator

from chatsky.core.service.extra import BeforeHandler, AfterHandler
from chatsky.core.script_function import AnyCondition
from chatsky.core.service.types import (
    ComponentExecutionState,
    ExtraHandlerType,
    ExtraHandlerFunction,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.core.context import Context


class PipelineComponent(abc.ABC, BaseModel, extra="forbid", arbitrary_types_allowed=True):
    """
    Base class for a single task processed by :py:class:`.Pipeline`.
    """

    before_handler: BeforeHandler = Field(default_factory=BeforeHandler)
    """
    :py:class:`~.BeforeHandler`, associated with this component.
    """
    after_handler: AfterHandler = Field(default_factory=AfterHandler)
    """
    :py:class:`~.AfterHandler`, associated with this component.
    """
    timeout: Optional[float] = None
    """
    Maximum component execution time (in seconds),
    if it exceeds this time, it is interrupted.
    """
    concurrent: bool = False
    """
    Optional flag that indicates whether this component
    should be executed concurrently with adjacent concurrent components.
    """
    start_condition: AnyCondition = Field(default=True, validate_default=True)
    """
    :py:data:`~.AnyCondition` that is invoked before each component execution;
    component is executed only if it returns ``True``.
    """
    name: Optional[str] = None
    """
    Name of the component. Defaults to :py:attr:`.computed_name` potentially modified by
    :py:func:`~chatsky.core.utils.rename_component_incrementing`.

    See :py:meth:`.validate_name` for rules.
    """
    path: Optional[str] = None
    """
    Separated by dots path to component, is universally unique.
    """

    @field_validator("name")
    @classmethod
    def validate_name(cls, name: str):
        """
        Validate this component's name:

        Name cannot be empty or contain "." or "#".

        :raises ValueError: If name failed validation.
        """
        if name is not None:
            if name == "":
                raise ValueError("Name cannot be blank.")
            if "." in name:
                raise ValueError(f"Name cannot contain '.': {name!r}.")
            if "#" in name:
                raise ValueError(f"Name cannot contain '#': {name!r}.")

        return name

    def _set_state(self, ctx: Context, value: ComponentExecutionState):
        """
        Method for component runtime state setting, state is preserved in :py:attr:`.Context.framework_data`.

        :param ctx: :py:class:`.Context` to keep state in.
        :param value: State to set.
        """
        ctx.framework_data.service_states[self.path].execution_status = value

    def get_state(self, ctx: Context) -> ComponentExecutionState:
        """
        Method for component runtime state getting, state is preserved in :py:attr:`.Context.framework_data`.

        :param ctx: :py:class:`~.Context` to get state from.
        :return: :py:class:`.ComponentExecutionState` of this service.
        """
        return ctx.framework_data.service_states[self.path].execution_status

    @abc.abstractmethod
    async def run_component(self, ctx: Context) -> Optional[ComponentExecutionState]:
        """
        Run this component.

        :param ctx: Current dialog :py:class:`~.Context`.
        """
        raise NotImplementedError

    @property
    def computed_name(self) -> str:
        """
        Default name that is used if :py:attr:`~.PipelineComponent.name` is not defined.

        In case two components in a :py:class:`~chatsky.core.service.group.ServiceGroup` have the same
        :py:attr:`.computed_name` an incrementing number is appended to the name.
        """
        return "noname_service"

    async def _run(self, ctx: Context) -> None:
        """
        A method for running a pipeline component. Executes extra handlers before and after execution,
        launches :py:meth:`.run_component` method. This method is run after the component's timeout is set (if needed).

        :param ctx: Current dialog :py:class:`~.Context`.
        """

        async def _inner_run():
            if await self.start_condition(ctx):
                await self.before_handler(ctx, self)

                self._set_state(ctx, ComponentExecutionState.RUNNING)
                logger.debug(f"Running component {self.path!r}")
                result = await self.run_component(ctx)
                if isinstance(result, ComponentExecutionState):
                    self._set_state(ctx, result)
                else:
                    self._set_state(ctx, ComponentExecutionState.FINISHED)

                await self.after_handler(ctx, self)
            else:
                self._set_state(ctx, ComponentExecutionState.NOT_RUN)

        try:
            await asyncio.wait_for(_inner_run(), timeout=self.timeout)
        except Exception as exc:
            self._set_state(ctx, ComponentExecutionState.FAILED)
            logger.error(f"Service '{self.name}' execution failed!", exc_info=exc)
        finally:
            ctx.framework_data.service_states[self.path].finished_event.set()

    async def __call__(self, ctx: Context) -> None:
        """
        A method for calling pipeline components.
        It sets up timeout and executes it using :py:meth:`_run` method.

        :param ctx: Current dialog :py:class:`~.Context`.
        :return: ``None``
        """
        await self._run(ctx)

    def add_extra_handler(self, extra_handler_type: ExtraHandlerType, extra_handler: ExtraHandlerFunction):
        """
        Add extra handler to this component.

        :param extra_handler_type: A type of extra handler to add (before or after).
        :param extra_handler: Function to add to the component as an extra handler.
        """
        if extra_handler_type == ExtraHandlerType.BEFORE:
            target = self.before_handler
        elif extra_handler_type == ExtraHandlerType.AFTER:
            target = self.after_handler
        else:
            raise ValueError(f"Unrecognized ExtraHandlerType: {extra_handler_type}")
        target.functions.append(extra_handler)



"""
This module defines core feature of the Chatsky framework.
"""

from chatsky.core.context import Context
from chatsky.core.message import (
    Message,
    MessageInitTypes,
    Attachment,
    CallbackQuery,
    Location,
    Contact,
    Invoice,
    PollOption,
    Poll,
    DataAttachment,
    Audio,
    Video,
    Animation,
    Image,
    Sticker,
    Document,
    VoiceMessage,
    VideoMessage,
    MediaGroup,
)
from chatsky.core.pipeline import Pipeline
from chatsky.core.script import Node, Flow, Script
from chatsky.core.script_function import BaseCondition, BaseResponse, BaseDestination, BaseProcessing, BasePriority
from chatsky.core.script_function import AnyCondition, AnyResponse, AnyDestination, AnyPriority
from chatsky.core.transition import Transition
from chatsky.core.node_label import NodeLabel, NodeLabelInitTypes, AbsoluteNodeLabel, AbsoluteNodeLabelInitTypes
from chatsky.core.script import GLOBAL, LOCAL, RESPONSE, TRANSITIONS, MISC, PRE_RESPONSE, PRE_TRANSITION



"""
Common
------
This module contains several functions which are used to run demonstrations in tutorials.
"""

from os import getenv
from typing import Tuple, Iterable
from uuid import uuid4

from chatsky.core import Message, Pipeline
from chatsky.core.message import MessageInitTypes


def is_interactive_mode() -> bool:  # pragma: no cover
    """
    Checking whether the tutorial code should be run in interactive mode.

    :return: `True` if it's being executed by Jupyter kernel and DISABLE_INTERACTIVE_MODE env variable isn't set,
        `False` otherwise.
    """

    shell = None
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
    finally:
        return shell != "ZMQInteractiveShell" and getenv("DISABLE_INTERACTIVE_MODE") is None


def check_happy_path(
    pipeline: Pipeline,
    happy_path: Iterable[Tuple[MessageInitTypes, MessageInitTypes]],
    *,
    response_comparator=Message.__eq__,
    printout: bool = False,
):
    """
    Running tutorial with provided pipeline for provided requests, comparing responses with correct expected responses.

    :param pipeline: The Pipeline instance, that will be used for checking.
    :param happy_path: A tuple of (request, response) tuples, so-called happy path,
        its requests are passed to pipeline and the pipeline responses are compared to its responses.
    :param response_comparator:
        Function that checks reference response (first argument) with the actual response (second argument).
        Defaults to ``Message.__eq__``.
    :param printout: Whether to print the requests/responses during iteration.
    """
    ctx_id = uuid4()  # get random ID for current context
    for step_id, (request_raw, reference_response_raw) in enumerate(happy_path):

        request = Message.model_validate(request_raw)
        reference_response = Message.model_validate(reference_response_raw)
        if printout:
            print(f"USER: {request}")

        ctx = pipeline(request, ctx_id)

        actual_response = ctx.last_response
        if printout:
            print(f"BOT : {actual_response}")

        if not response_comparator(reference_response, actual_response):
            raise AssertionError(
                f"""check_happy_path failed
step id: {step_id}
reference response: {reference_response}
actual response: {actual_response}
"""
            )



from chatsky.conditions.standard import (
    ExactMatch,
    HasText,
    Regexp,
    Any,
    All,
    Negation,
    CheckLastLabels,
    Not,
    HasCallbackQuery,
)
from chatsky.conditions.slots import SlotsExtracted
from chatsky.conditions.service import ServiceFinished



"""
Service
-------
The Service module contains the :py:class:`.Service` class which represents a single task.

Pipeline consists of services and service groups.
Service is an atomic part of a pipeline.
"""

from __future__ import annotations
import logging
import inspect
from typing import Any, Optional, Callable, Union
from typing_extensions import TypeAlias, Annotated
from pydantic import model_validator, Field

from chatsky.core.context import Context
from chatsky.core.script_function import BaseProcessing, AnyCondition
from chatsky.core.service.component import PipelineComponent
from .extra import BeforeHandler, AfterHandler
from chatsky.utils.devel import wrap_sync_function_in_async

logger = logging.getLogger(__name__)


class Service(PipelineComponent):
    """
    This class represents a service.
    """

    handler: Union[BaseProcessing, Callable[[Context], None]] = None
    """
    Function that represents this service.
    """
    # Repeating inherited fields for better documentation.
    before_handler: BeforeHandler = Field(default_factory=BeforeHandler)
    after_handler: AfterHandler = Field(default_factory=AfterHandler)
    timeout: Optional[float] = None
    concurrent: bool = False
    start_condition: AnyCondition = Field(default=True, validate_default=True)
    name: Optional[str] = None
    path: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def handler_validator(cls, data: Any):
        """
        Add support for initializing from a `Callable` or `BaseProcessing`.
        """
        if inspect.isfunction(data) or isinstance(data, BaseProcessing):
            return {"handler": data}
        return data

    async def call(self, ctx: Context) -> None:
        """
        A placeholder method which the user can redefine in their own derivative of :py:class:`.Service`.
        This allows direct access to the ``self`` object and all its fields.

        :param ctx: Current dialog context.
        """
        if self.handler is None:
            raise NotImplementedError(
                f"Received {self.__class__.__name__} object, which has it's 'handler' == 'None',"
                f" while also not defining it's own 'call()' method."
            )
        await wrap_sync_function_in_async(self.handler, ctx)

    async def run_component(self, ctx: Context) -> None:
        """
        Method for running this service.

        :param ctx: Current dialog context.
        :return: `None`
        """
        await wrap_sync_function_in_async(self.call, ctx)

    @property
    def computed_name(self) -> str:
        """
        Return name of the handler or name of this class if handler is empty.
        """
        if inspect.isfunction(self.handler):
            return self.handler.__name__
        elif self.handler is None:
            return self.__class__.__name__
        else:
            return self.handler.__class__.__name__


def to_service(
    before_handler: BeforeHandler = None,
    after_handler: AfterHandler = None,
    timeout: Optional[int] = None,
    concurrent: bool = False,
    start_condition: AnyCondition = True,
    name: Optional[str] = None,
):
    """
    Return a function decorator that returns a :py:class:`Service` with the
    given parameters.
    """
    before_handler = BeforeHandler() if before_handler is None else before_handler
    after_handler = AfterHandler() if after_handler is None else after_handler

    def inner(handler: Union[BaseProcessing, Callable]) -> Service:
        return Service(
            handler=handler,
            before_handler=before_handler,
            after_handler=after_handler,
            timeout=timeout,
            concurrent=concurrent,
            start_condition=start_condition,
            name=name,
        )

    return inner


ServiceInitTypes: TypeAlias = Union[
    Service,
    Annotated[dict, "dict following the Service data model"],
    Annotated[Callable, "handler for the service"],
    Annotated[BaseProcessing, "handler for the service"],
]
"""Types that :py:class:`~.Service` can be validated from."""



"""
Standard Destinations
---------------------
This module provides basic destinations.

- :py:class:`FromHistory`, :py:class:`Current` and :py:class:`Previous` -- history-based destinations;
- :py:class:`Start` and :py:class:`Fallback` -- config-based destinations;
- :py:class:`Forward` and :py:class:`Backward` -- script-based destinations.
"""

from __future__ import annotations

from pydantic import Field

from chatsky.core.context import get_last_index, Context
from chatsky.core.node_label import NodeLabelInitTypes, AbsoluteNodeLabel
from chatsky.core.script_function import BaseDestination


class FromHistory(BaseDestination):
    """
    Return label of the node located at a certain position in the label history.
    """

    position: int = Field(le=-1)
    """
    Position of the label in label history.

    Should be negative:

    - ``-1`` refers to the current node (same as ``ctx.last_label``);
    - ``-2`` -- to the previous node.
    """

    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        index = get_last_index(ctx.labels)
        shifted_index = index + self.position + 1
        result = ctx.labels.get(shifted_index)
        if result is None:
            raise KeyError(
                f"No label with index {shifted_index!r}. "
                f"Current label index: {index!r}; FromHistory.position: {self.position!r}."
            )
        return result


class Current(FromHistory):
    """
    Return label of the current node.
    """

    position: int = -1
    """Position is set to ``-1`` to get the last node."""


class Previous(FromHistory):
    """
    Return label of the previous node.
    """

    position: int = -2
    """Position is set to ``-2`` to get the second to last node."""


class Start(BaseDestination):
    """
    Return :py:attr:`~chatsky.core.pipeline.Pipeline.start_label`.
    """

    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        return ctx.pipeline.start_label


class Fallback(BaseDestination):
    """
    Return :py:attr:`~chatsky.core.pipeline.Pipeline.fallback_label`.
    """

    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        return ctx.pipeline.fallback_label


def get_next_node_in_flow(
    node_label: AbsoluteNodeLabel,
    ctx: Context,
    *,
    increment: bool = True,
    loop: bool = False,
) -> AbsoluteNodeLabel:
    """
    Function that returns node label of a node in the same flow after shifting the index.

    :param node_label: Label of the node to shift from.
    :param ctx: Dialog context.
    :param increment: If it is `True`, label index is incremented by `1`,
        otherwise it is decreased by `1`.
    :param loop: If it is `True` the iteration over the label list is going cyclically
        (i.e. Backward in the first node returns the last node).
    :return: The tuple that consists of `(flow_label, label, priority)`.
        If fallback is executed `(flow_fallback_label, fallback_label, priority)` are returned.
    """
    node_label = AbsoluteNodeLabel.model_validate(node_label, context={"ctx": ctx})
    node_keys = list(ctx.pipeline.script.get_flow(node_label.flow_name).nodes.keys())

    node_index = node_keys.index(node_label.node_name)
    node_index = node_index + 1 if increment else node_index - 1
    if not (loop or (0 <= node_index < len(node_keys))):
        raise IndexError(
            f"Node index {node_index!r} out of range for node_keys: {node_keys!r}. Consider using the `loop` flag."
        )
    node_index %= len(node_keys)

    return AbsoluteNodeLabel(flow_name=node_label.flow_name, node_name=node_keys[node_index])


class Forward(BaseDestination):
    """
    Return the next node relative to the current node in the current flow.
    """

    loop: bool = False
    """
    Whether to return the first node of the flow if the current node is the last one.
    Otherwise and exception is raised (and transition is considered unsuccessful).
    """

    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        return get_next_node_in_flow(ctx.last_label, ctx, increment=True, loop=self.loop)


class Backward(BaseDestination):
    """
    Return the previous node relative to the current node in the current flow.
    """

    loop: bool = False
    """
    Whether to return the last node of the flow if the current node is the first one.
    Otherwise and exception is raised (and transition is considered unsuccessful).
    """

    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        return get_next_node_in_flow(ctx.last_label, ctx, increment=False, loop=self.loop)



"""
Node Label
----------
This module defines classes for addressing nodes.
"""

from __future__ import annotations

from typing import Optional, Union, Tuple, List, TYPE_CHECKING
from typing_extensions import TypeAlias, Annotated

from pydantic import BaseModel, model_validator, ValidationInfo

if TYPE_CHECKING:
    from chatsky.core.context import Context


def _get_current_flow_name(ctx: Context) -> str:
    """Get flow name of the current node from context."""
    current_node = ctx.last_label
    return current_node.flow_name


class NodeLabel(BaseModel, frozen=True):
    """
    A label for a node. (a way to address a specific node in the script)

    Can be relative if :py:attr:`flow_name` is ``None``:
    such ``NodeLabel`` will reference a node with the name :py:attr:`node_name`
    in the current flow.
    """

    flow_name: Optional[str] = None
    """
    Name of the flow in the script.
    Can be ``None`` in which case this is inherited from the :py:attr:`.Context.current_node`.
    """
    node_name: str
    """
    Name of the node in the flow.
    """

    @model_validator(mode="before")
    @classmethod
    def validate_from_str_or_tuple(cls, data, info: ValidationInfo):
        """
        Allow instantiating of this class from:

        - A single string (node name). Also attempt to get the current flow name from context.
        - A tuple or list of two strings (flow and node name).
        """
        if isinstance(data, str):
            flow_name = None
            context = info.context
            if isinstance(context, dict):
                flow_name = _get_current_flow_name(context.get("ctx"))
            return {"flow_name": flow_name, "node_name": data}
        elif isinstance(data, (tuple, list)):
            if len(data) == 2 and isinstance(data[0], str) and isinstance(data[1], str):
                return {"flow_name": data[0], "node_name": data[1]}
            else:
                raise ValueError(
                    f"Cannot validate NodeLabel from {data!r}: {type(data).__name__} should contain 2 strings."
                )
        return data


NodeLabelInitTypes: TypeAlias = Union[
    NodeLabel,
    Annotated[str, "node_name, flow name equal to current flow's name"],
    Tuple[Annotated[str, "flow_name"], Annotated[str, "node_name"]],
    Annotated[List[str], "list of two strings (flow_name and node_name)"],
    Annotated[dict, "dict following the NodeLabel data model"],
]
"""Types that :py:class:`~.NodeLabel` can be validated from."""


class AbsoluteNodeLabel(NodeLabel):
    """
    A label for a node. (a way to address a specific node in the script)
    """

    flow_name: str
    """
    Name of the flow in the script.
    """
    node_name: str
    """
    Name of the node in the flow.
    """

    @model_validator(mode="before")
    @classmethod
    def validate_from_node_label(cls, data, info: ValidationInfo):
        """
        Allow instantiating of this class from :py:class:`NodeLabel`.

        Attempt to get the current flow name from context if :py:attr:`NodeLabel.flow_name` is empty.
        """
        if isinstance(data, NodeLabel):
            flow_name = data.flow_name
            if flow_name is None:
                context = info.context
                if isinstance(context, dict):
                    flow_name = _get_current_flow_name(context.get("ctx"))
            return {"flow_name": flow_name, "node_name": data.node_name}
        return data

    @model_validator(mode="after")
    def check_node_exists(self, info: ValidationInfo):
        """
        Validate node exists in the script.
        """
        context = info.context
        if isinstance(context, dict):
            ctx: Context = info.context.get("ctx")
            if ctx is not None:
                script = ctx.pipeline.script

                node = script.get_node(self)
                if node is None:
                    raise ValueError(f"Cannot find node {self!r} in script.")
        return self


AbsoluteNodeLabelInitTypes: TypeAlias = Union[
    AbsoluteNodeLabel,
    NodeLabel,
    Tuple[Annotated[str, "flow_name"], Annotated[str, "node_name"]],
    Annotated[List[str], "list of two strings (flow_name and node_name)"],
    Annotated[dict, "dict following the AbsoluteNodeLabel data model"],
]
"""Types that :py:class:`~.AbsoluteNodeLabel` can be validated from."""



"""
Script
------
The Script module provides a set of `pydantic` models for representing the dialog graph.

These models are used by :py:class:`~chatsky.core.service.Actor` to define the conversation flow,
and to determine the appropriate response based on the user's input and the current state of the conversation.
"""

# %%
from __future__ import annotations
import logging
from typing import List, Optional, Dict

from pydantic import BaseModel, Field, AliasChoices

from chatsky.core.script_function import AnyResponse, BaseProcessing
from chatsky.core.node_label import AbsoluteNodeLabel
from chatsky.core.transition import Transition

logger = logging.getLogger(__name__)


class Node(BaseModel, extra="forbid"):
    """
    Node is a basic element of the dialog graph.

    Usually used to represent a specific state of a conversation.
    """

    transitions: List[Transition] = Field(
        validation_alias=AliasChoices("transitions", "TRANSITIONS"), default_factory=list
    )
    """List of transitions possible from this node."""
    response: Optional[AnyResponse] = Field(validation_alias=AliasChoices("response", "RESPONSE"), default=None)
    """Response produced when this node is entered."""
    pre_transition: Dict[str, BaseProcessing] = Field(
        validation_alias=AliasChoices("pre_transition", "PRE_TRANSITION"), default_factory=dict
    )
    """
    A dictionary of :py:class:`.BaseProcessing` functions that are executed before transitions are processed.
    Keys of the dictionary act as names for the processing functions.
    """
    pre_response: Dict[str, BaseProcessing] = Field(
        validation_alias=AliasChoices("pre_response", "PRE_RESPONSE"), default_factory=dict
    )
    """
    A dictionary of :py:class:`.BaseProcessing` functions that are executed before response is processed.
    Keys of the dictionary act as names for the processing functions.
    """
    misc: dict = Field(validation_alias=AliasChoices("misc", "MISC"), default_factory=dict)
    """
    A dictionary that is used to store metadata about the node.

    Can be accessed at runtime via :py:attr:`~chatsky.core.context.Context.current_node`.
    """

    def inherit_from_other(self, other: Node):
        """
        Inherit properties from another node into this one:

        - Extend ``self.transitions`` with :py:attr:`transitions` of the other node;
        - Replace response with ``other.response`` if ``self.response`` is ``None``;
        - Dictionaries (:py:attr:`pre_transition`, :py:attr:`pre_response` and :py:attr:`misc`)
          are appended to this node's dictionaries except for the repeating keys.
          For example, ``inherit_from_other({1: 1, 3: 3}, {1: 0, 2: 2}) == {1: 1, 3: 3, 2: 2}``.

        Basically, only non-conflicting properties of ``other`` are inherited.
        """

        def merge_dicts(first: dict, second: dict):
            first.update({k: v for k, v in second.items() if k not in first})

        self.transitions.extend(other.transitions)
        if self.response is None:
            self.response = other.response
        merge_dicts(self.pre_transition, other.pre_transition)
        merge_dicts(self.pre_response, other.pre_response)
        merge_dicts(self.misc, other.misc)
        return self


class Flow(BaseModel, extra="allow"):
    """
    Flow is a collection of nodes.
    This is used to group them by a specific purpose.
    """

    local_node: Node = Field(
        validation_alias=AliasChoices("local", "LOCAL", "local_node", "LOCAL_NODE"), default_factory=Node
    )
    """
    Node from which all other nodes in this Flow inherit properties
    according to :py:meth:`Node.inherit_from_other`.
    """
    __pydantic_extra__: Dict[str, Node]

    @property
    def nodes(self) -> Dict[str, Node]:
        """
        A dictionary of all non-local nodes in this flow.

        Keys in the dictionary acts as names for the nodes.
        """
        return self.__pydantic_extra__

    def get_node(self, name: str) -> Optional[Node]:
        """
        Get node with the ``name``.

        :return: Node or ``None`` if it doesn't exist.
        """
        return self.nodes.get(name)


class Script(BaseModel, extra="allow"):
    """
    A script is a collection of nodes.
    It represents an entire dialog graph.
    """

    global_node: Node = Field(
        validation_alias=AliasChoices("global", "GLOBAL", "global_node", "GLOBAL_NODE"), default_factory=Node
    )
    """
    Node from which all other nodes in this Script inherit properties
    according to :py:meth:`Node.inherit_from_other`.
    """
    __pydantic_extra__: Dict[str, Flow]

    @property
    def flows(self) -> Dict[str, Flow]:
        """
        A dictionary of all flows in this script.

        Keys in the dictionary acts as names for the flows.
        """
        return self.__pydantic_extra__

    def get_flow(self, name: str) -> Optional[Flow]:
        """
        Get flow with the ``name``.

        :return: Flow or ``None`` if it doesn't exist.
        """
        return self.flows.get(name)

    def get_node(self, label: AbsoluteNodeLabel) -> Optional[Node]:
        """
        Get node with the ``label``.

        :return: Node or ``None`` if it doesn't exist.
        """
        flow = self.get_flow(label.flow_name)
        if flow is None:
            return None
        return flow.get_node(label.node_name)

    def get_inherited_node(self, label: AbsoluteNodeLabel) -> Optional[Node]:
        """
        Return a new node that inherits (using :py:meth:`Node.inherit_from_other`)
        properties from :py:class:`Node`, :py:attr:`Flow.local_node`
        and :py:attr:`Script.global_node` (in that order).

        Flow and node are determined by ``label``.

        This is essentially a copy of the node specified by ``label``,
        that inherits properties from ``local_node`` and ``global_node``.

        :return: A new node or ``None`` if it doesn't exist.
        """
        flow = self.get_flow(label.flow_name)
        if flow is None:
            return None
        node = flow.get_node(label.node_name)
        if node is None:
            return None

        inheritant_node = Node()

        return (
            inheritant_node.inherit_from_other(node)
            .inherit_from_other(flow.local_node)
            .inherit_from_other(self.global_node)
        )


GLOBAL = "GLOBAL"
"""Key for :py:attr:`~chatsky.core.script.Script.global_node`."""
LOCAL = "LOCAL"
"""Key for :py:attr:`~chatsky.core.script.Flow.local_node`."""
TRANSITIONS = "TRANSITIONS"
"""Key for :py:attr:`~chatsky.core.script.Node.transitions`."""
RESPONSE = "RESPONSE"
"""Key for :py:attr:`~chatsky.core.script.Node.response`."""
MISC = "MISC"
"""Key for :py:attr:`~chatsky.core.script.Node.misc`."""
PRE_RESPONSE = "PRE_RESPONSE"
"""Key for :py:attr:`~chatsky.core.script.Node.pre_response`."""
PRE_TRANSITION = "PRE_TRANSITION"
"""Key for :py:attr:`~chatsky.core.script.Node.pre_transition`."""



"""
Extra Handler
-------------
Extra handlers are functions that are executed before or after a
:py:class:`~chatsky.core.service.component.PipelineComponent`.
"""

from __future__ import annotations
import asyncio
import logging
import inspect
import abc
from typing import Optional, List, Any, ClassVar, Union, Callable, TYPE_CHECKING
from typing_extensions import Annotated, TypeAlias

from pydantic import BaseModel, model_validator, Field

from chatsky.core.context import Context

from chatsky.utils.devel.async_helpers import wrap_sync_function_in_async
from chatsky.core.service.types import (
    ExtraHandlerType,
    ExtraHandlerFunction,
    ExtraHandlerRuntimeInfo,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.core.service import PipelineComponent


class ComponentExtraHandler(BaseModel, abc.ABC, extra="forbid", arbitrary_types_allowed=True):
    """
    Class, representing an extra handler for pipeline components.

    A component extra handler is a set of functions, attached to pipeline component (before or after it).
    Extra handlers should execute supportive tasks (like time or resources measurement, minor data transformations).
    """

    functions: List[ExtraHandlerFunction] = Field(default_factory=list)
    """
    A list or instance of :py:data:`~.ExtraHandlerFunction`.
    """
    stage: ClassVar[ExtraHandlerType]
    """
    An :py:class:`~.ExtraHandlerType`, specifying whether this handler will
    be executed before or after pipeline component.
    """
    timeout: Optional[float] = None
    """
    Maximum component execution time (in seconds),
    if it exceeds this time, it is interrupted.
    """
    concurrent: bool = False
    """
    A flag that indicates whether the extra handler's functions
    should be executed concurrently. False by default.
    """

    @model_validator(mode="before")
    @classmethod
    def functions_validator(cls, data: Any):
        """
        Add support for initializing from a `Callable` or List[`Callable`].
        Casts `functions` to `list` if it's not already.
        """
        if isinstance(data, list):
            result = {"functions": data}
        elif callable(data):
            result = {"functions": [data]}
        else:
            result = data

        if isinstance(result, dict):
            if ("functions" in result) and (not isinstance(result["functions"], list)):
                result["functions"] = [result["functions"]]
        return result

    async def _run_function(self, func: ExtraHandlerFunction, ctx: Context, component: PipelineComponent):
        handler_params = len(inspect.signature(func).parameters)
        if handler_params == 1:
            await wrap_sync_function_in_async(func, ctx)
        elif handler_params == 2:
            extra_handler_runtime_info = ExtraHandlerRuntimeInfo(stage=self.stage, component=component)
            await wrap_sync_function_in_async(func, ctx, extra_handler_runtime_info)
        else:
            raise Exception(
                f"Too many parameters required for component {component.name} {self.stage}"
                f" wrapper handler '{func.__name__}': {handler_params}!"
            )

    async def _run(self, ctx: Context, component_info: PipelineComponent):
        """
        Method for executing one of the extra handler functions (before or after).
        If the function is not set, nothing happens.

        :param ctx: current dialog context.
        :param component_info: associated component's `self` object.
        :return: `None`
        """

        if self.concurrent:
            await asyncio.gather(*[self._run_function(func, ctx, component_info) for func in self.functions])
        else:
            for func in self.functions:
                await self._run_function(func, ctx, component_info)

    async def __call__(self, ctx: Context, component_info: PipelineComponent):
        """
        A method for calling an extra handler.
        It sets up a timeout and executes it using `_run` method.

        :param ctx: (required) Current dialog `Context`.
        :param component_info: associated component's `self` object.
        """
        task = asyncio.create_task(self._run(ctx, component_info))
        await asyncio.wait_for(task, timeout=self.timeout)


class BeforeHandler(ComponentExtraHandler):
    """
    A handler for extra functions that are executed before the component's main function.

    :param functions: A list of callables that will be executed
        before the component's main function.
    :type functions: List[ExtraHandlerFunction]
    :param timeout: Optional timeout for the execution of the extra functions, in
        seconds.
    :param concurrent: Optional flag that indicates whether the extra functions
        should be executed concurrently. False by default.
    """

    stage: ClassVar[ExtraHandlerType] = ExtraHandlerType.BEFORE


class AfterHandler(ComponentExtraHandler):
    """
    A handler for extra functions that are executed after the component's main function.

    :param functions: A list of callables that will be executed
        after the component's main function.
    :type functions: List[ExtraHandlerFunction]
    :param timeout: Optional timeout for the execution of the extra functions, in
        seconds.
    :param concurrent: Optional flag that indicates whether the extra functions
        should be executed concurrently. False by default.
    """

    stage: ClassVar[ExtraHandlerType] = ExtraHandlerType.AFTER


ComponentExtraHandlerInitTypes: TypeAlias = Union[
    ComponentExtraHandler,
    Annotated[dict, "dict following the ComponentExtraHandler data model"],
    Annotated[Callable, "a singular function for the extra handler"],
    Annotated[List[Callable], "functions for the extra handler"],
]
"""Types that :py:class:`~.ComponentExtraHandler` can be validated from."""



from chatsky.messengers.common import (
    MessengerInterface,
    MessengerInterfaceWithAttachments,
    PollingMessengerInterface,
    CallbackMessengerInterface,
)
from chatsky.messengers.telegram import LongpollingInterface as TelegramInterface
from chatsky.messengers.console import CLIMessengerInterface



"""
Types
-----
This module defines type aliases used throughout the ``Core.Service`` module.

The classes and special types in this module can include data models,
data structures, and other types that are defined for type hinting.
"""

from __future__ import annotations
from enum import unique, Enum
from typing import Callable, Union, Awaitable, Optional, Any, Protocol, Hashable, TYPE_CHECKING
from typing_extensions import TypeAlias
from pydantic import BaseModel


if TYPE_CHECKING:
    from chatsky.core import Context, Message
    from chatsky.core.service import PipelineComponent


class PipelineRunnerFunction(Protocol):
    """
    Protocol for pipeline running.
    """

    def __call__(
        self, message: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Awaitable[Context]:
        """
        :param message: User request for pipeline to process.
        :param ctx_id:
            ID of the context that the new request belongs to.
            Optional, None by default.
            If set to `None`, a new context will be created with `message` being the first request.
        :param update_ctx_misc:
            Dictionary to be passed as an argument to `ctx.misc.update`.
            This argument can be used to store values in the `misc` dictionary before anything else runs.
            Optional; None by default.
            If set to `None`, `ctx.misc.update` will not be called.
        :return:
            Context instance that pipeline processed.
            The context instance has the id of `ctx_id`.
            If `ctx_id` is `None`, context instance has an id generated with `uuid.uuid4`.
        """


@unique
class ComponentExecutionState(str, Enum):
    """
    Enum, representing pipeline component execution state.
    These states are stored in :py:attr:`~chatsky.core.context.FrameworkData.service_states`,
    that should always be requested with ``NOT_RUN`` being default fallback.
    Following states are supported:

    - NOT_RUN: component has not been executed yet (the default one),
    - RUNNING: component is currently being executed,
    - FINISHED: component executed successfully,
    - FAILED: component execution failed.
    """

    NOT_RUN = "NOT_RUN"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


@unique
class ExtraHandlerType(str, Enum):
    """
    Enum, representing extra handler execution stage: before or after the wrapped function.
    The following types are supported:

    - BEFORE: extra handler function called before component,
    - AFTER: extra handler function called after component.
    """

    BEFORE = "BEFORE"
    AFTER = "AFTER"


ExtraHandlerConditionFunction: TypeAlias = Callable[[str], bool]
"""
A function type used when adding an extra handler to a service group to determine
whether extra handler should be added to components of the service group.

Accepts a single argument - path of a subcomponent in a group.
Return bool - whether to add extra handler to the subcomponent.
"""


ExtraHandlerFunction: TypeAlias = Union[
    Callable[["Context"], Any],
    Callable[["Context", "ExtraHandlerRuntimeInfo"], Any],
]
"""
A function type for creating extra handler (before and after functions).
Can accept current dialog context and current extra handler info.
"""


class ExtraHandlerRuntimeInfo(BaseModel):
    """
    Information passed to :py:data:`.ExtraHandlerFunction`.
    """

    stage: ExtraHandlerType
    """
    :py:class:`.ExtraHandlerType` -- either "BEFORE" or "AFTER".
    """
    component: PipelineComponent
    """
    Component object that called the extra handler.
    """



"""
Toy script
----------
This module contains a simple script and a dialog which are used
in tutorials.
"""

from chatsky.conditions import ExactMatch
from chatsky.core import TRANSITIONS, RESPONSE, Transition as Tr

TOY_SCRIPT = {
    "greeting_flow": {
        "start_node": {
            TRANSITIONS: [Tr(dst="node1", cnd=ExactMatch("Hi"))],
        },
        "node1": {
            RESPONSE: "Hi, how are you?",
            TRANSITIONS: [Tr(dst="node2", cnd=ExactMatch("i'm fine, how are you?"))],
        },
        "node2": {
            RESPONSE: "Good. What do you want to talk about?",
            TRANSITIONS: [Tr(dst="node3", cnd=ExactMatch("Let's talk about music."))],
        },
        "node3": {
            RESPONSE: "Sorry, I can not talk about music now.",
            TRANSITIONS: [Tr(dst="node4", cnd=ExactMatch("Ok, goodbye."))],
        },
        "node4": {RESPONSE: "bye", TRANSITIONS: [Tr(dst="node1", cnd=ExactMatch("Hi"))]},
        "fallback_node": {
            RESPONSE: "Ooops",
            TRANSITIONS: [Tr(dst="node1", cnd=ExactMatch("Hi"))],
        },
    }
}
"""
An example of a simple script.

:meta hide-value:
"""

TOY_SCRIPT_KWARGS = {
    "script": TOY_SCRIPT,
    "start_label": ("greeting_flow", "start_node"),
    "fallback_label": ("greeting_flow", "fallback_node"),
}
"""
# There should be a better description of this
Keyword arguments to pass to :py:meth:`~chatsky.core.pipeline.Pipeline` in order to
use :py:data:`~.TOY_SCRIPT`:

.. code-block::

    Pipeline(**TOY_SCRIPT_KWARGS, context_storage=...)

:meta hide-value:
"""

HAPPY_PATH = (
    ("Hi", "Hi, how are you?"),
    ("i'm fine, how are you?", "Good. What do you want to talk about?"),
    ("Let's talk about music.", "Sorry, I can not talk about music now."),
    ("Ok, goodbye.", "bye"),
    ("Hi", "Hi, how are you?"),
)
"""
An example of a simple dialog.

:meta hide-value:
"""

MULTIFLOW_SCRIPT = {
    "root": {
        "start": {
            RESPONSE: "Hi",
            TRANSITIONS: [
                Tr(dst=("small_talk", "ask_some_questions"), cnd=ExactMatch("hi")),
                Tr(dst=("animals", "have_pets"), cnd=ExactMatch("i like animals")),
                Tr(dst=("animals", "like_animals"), cnd=ExactMatch("let's talk about animals")),
                Tr(dst=("news", "what_news"), cnd=ExactMatch("let's talk about news")),
            ],
        },
        "fallback": {RESPONSE: "Oops", TRANSITIONS: [Tr(dst="start")]},
    },
    "animals": {
        "have_pets": {
            RESPONSE: "do you have pets?",
            TRANSITIONS: [Tr(dst="what_animal", cnd=ExactMatch("yes"))],
        },
        "like_animals": {
            RESPONSE: "do you like it?",
            TRANSITIONS: [Tr(dst="what_animal", cnd=ExactMatch("yes"))],
        },
        "what_animal": {
            RESPONSE: "what animals do you have?",
            TRANSITIONS: [
                Tr(dst="ask_about_color", cnd=ExactMatch("bird")),
                Tr(dst="ask_about_breed", cnd=ExactMatch("dog")),
            ],
        },
        "ask_about_color": {RESPONSE: "what color is it"},
        "ask_about_breed": {
            RESPONSE: "what is this breed?",
            TRANSITIONS: [
                Tr(dst="ask_about_breed", cnd=ExactMatch("pereat")),
                Tr(dst="tell_fact_about_breed", cnd=ExactMatch("bulldog")),
                Tr(dst="ask_about_training", cnd=ExactMatch("I don't know")),
            ],
        },
        "tell_fact_about_breed": {
            RESPONSE: "Bulldogs appeared in England as specialized bull-baiting dogs. ",
        },
        "ask_about_training": {RESPONSE: "Do you train your dog? "},
    },
    "news": {
        "what_news": {
            RESPONSE: "what kind of news do you prefer?",
            TRANSITIONS: [
                Tr(dst="ask_about_science", cnd=ExactMatch("science")),
                Tr(dst="ask_about_sport", cnd=ExactMatch("sport")),
            ],
        },
        "ask_about_science": {
            RESPONSE: "i got news about science, do you want to hear?",
            TRANSITIONS: [
                Tr(dst="science_news", cnd=ExactMatch("yes")),
                Tr(dst=("small_talk", "ask_some_questions"), cnd=ExactMatch("let's change the topic")),
            ],
        },
        "science_news": {
            RESPONSE: "This is science news",
            TRANSITIONS: [
                Tr(dst="what_news", cnd=ExactMatch("ok")),
                Tr(dst=("small_talk", "ask_some_questions"), cnd=ExactMatch("let's change the topic")),
            ],
        },
        "ask_about_sport": {
            RESPONSE: "i got news about sport, do you want to hear?",
            TRANSITIONS: [
                Tr(dst="sport_news", cnd=ExactMatch("yes")),
                Tr(dst=("small_talk", "ask_some_questions"), cnd=ExactMatch("let's change the topic")),
            ],
        },
        "sport_news": {
            RESPONSE: "This is sport news",
            TRANSITIONS: [
                Tr(dst="what_news", cnd=ExactMatch("ok")),
                Tr(dst=("small_talk", "ask_some_questions"), cnd=ExactMatch("let's change the topic")),
            ],
        },
    },
    "small_talk": {
        "ask_some_questions": {
            RESPONSE: "how are you",
            TRANSITIONS: [
                Tr(dst="ask_talk_about", cnd=ExactMatch("fine")),
                Tr(dst=("animals", "like_animals"), cnd=ExactMatch("let's talk about animals")),
                Tr(dst=("news", "what_news"), cnd=ExactMatch("let's talk about news")),
            ],
        },
        "ask_talk_about": {
            RESPONSE: "what do you want to talk about",
            TRANSITIONS: [
                Tr(dst=("animals", "like_animals"), cnd=ExactMatch("dog")),
                Tr(dst=("news", "what_news"), cnd=ExactMatch("let's talk about news")),
            ],
        },
    },
}
"""
Simple dialog with multiple flows.

:meta hide-value:
"""

MULTIFLOW_REQUEST_OPTIONS = {
    "root": {
        "start": [
            "hi",
            "i like animals",
            "let's talk about animals",
        ],
        "fallback": [
            "to start",
        ],
    },
    "animals": {
        "have_pets": ["yes"],
        "like_animals": ["yes"],
        "what_animal": ["bird", "dog"],
        "ask_about_breed": ["pereat", "bulldog", "I don't know"],
    },
    "news": {
        "what_news": ["science", "sport"],
        "ask_about_science": ["yes", "let's change the topic"],
        "science_news": ["ok", "let's change the topic"],
        "ask_about_sport": ["yes", "let's change the topic"],
        "sport_news": ["ok", "let's change the topic"],
    },
    "small_talk": {
        "ask_some_questions": [
            "fine",
            "let's talk about animals",
            "let's talk about news",
        ],
        "ask_talk_about": ["dog", "let's talk about news"],
    },
}
"""
Request options for automated client requests for :py:data:`~.MULTIFLOW_SCRIPT`.

:meta hide-value:
"""



"""
Pipeline File Import
--------------------
This module introduces tools that allow importing Pipeline objects from
json/yaml files.

- :py:class:`JSONImporter` is a class that imports pipeline from files
- :py:func:`get_chatsky_objects` is a function that provides an index of objects commonly used in a Pipeline definition.
"""

from typing import Union, Optional, Any, List, Tuple
import importlib
import importlib.util
import importlib.machinery
import sys
import logging
from pathlib import Path
import json
from inspect import ismodule
from functools import reduce
from contextlib import contextmanager

from pydantic import JsonValue

try:
    import yaml

    yaml_available = True
except ImportError:
    yaml_available = False


logger = logging.getLogger(__name__)


class JSONImportError(Exception):
    """An exception for incorrect usage of :py:class:`JSONImporter`."""

    __notes__ = [
        "Read the guide on Pipeline import from file: "
        "https://deeppavlov.github.io/chatsky/user_guides/pipeline_import.html"
    ]


class JSONImporter:
    """
    Enables pipeline import from file.

    Since Pipeline and all its components are already pydantic ``BaseModel``,
    the only purpose of this class is to allow importing and instantiating arbitrary objects.

    Import is done by replacing strings of certain patterns with corresponding objects.
    This process is implemented in :py:meth:`resolve_string_reference`.

    Instantiating is done by replacing dictionaries where a single key is an imported object
    with an initialized object where arguments are specified by the dictionary values.
    This process is implemented in :py:meth:`replace_resolvable_objects` and
    :py:meth:`parse_args`.

    :param custom_dir: Path to the directory containing custom code available for import under the
        :py:attr:`CUSTOM_DIR_NAMESPACE_PREFIX`.
    """

    CHATSKY_NAMESPACE_PREFIX: str = "chatsky."
    """
    Prefix that indicates an import from the `chatsky` library.

    This class variable can be changed to allow using a different prefix.
    """
    CUSTOM_DIR_NAMESPACE_PREFIX: str = "custom."
    """
    Prefix that indicates an import from the custom directory.

    This class variable can be changed to allow using a different prefix.
    """
    EXTERNAL_LIB_NAMESPACE_PREFIX: str = "external:"
    """
    Prefix that indicates an import from any library.

    This class variable can be changed to allow using a different prefix.
    """

    def __init__(self, custom_dir: Union[str, Path]):
        self.custom_dir: Path = Path(custom_dir).absolute()
        self.custom_dir_location: str = str(self.custom_dir.parent)
        self.custom_dir_stem: str = str(self.custom_dir.stem)

    @staticmethod
    def is_resolvable(value: str) -> bool:
        """
        Check if ``value`` starts with any of the namespace prefixes:

        - :py:attr:`CHATSKY_NAMESPACE_PREFIX`;
        - :py:attr:`CUSTOM_DIR_NAMESPACE_PREFIX`;
        - :py:attr:`EXTERNAL_LIB_NAMESPACE_PREFIX`.

        :return: Whether the value should be resolved (starts with a namespace prefix).
        """
        return (
            value.startswith(JSONImporter.CHATSKY_NAMESPACE_PREFIX)
            or value.startswith(JSONImporter.CUSTOM_DIR_NAMESPACE_PREFIX)
            or value.startswith(JSONImporter.EXTERNAL_LIB_NAMESPACE_PREFIX)
        )

    @staticmethod
    @contextmanager
    def sys_path_append(path):
        """
        Append ``path`` to ``sys.path`` before yielding and
        restore ``sys.path`` to initial state after returning.
        """
        sys_path = sys.path.copy()
        sys.path.append(path)
        yield
        sys.path = sys_path

    @staticmethod
    def replace_prefix(string, old_prefix, new_prefix) -> str:
        """
        Replace ``old_prefix`` in ``string`` with ``new_prefix``.

        :raises ValueError: If the ``string`` does not begin with ``old_prefix``.
        :return: A new string with a new prefix.
        """
        if not string.startswith(old_prefix):
            raise ValueError(f"String {string!r} does not start with {old_prefix!r}")
        return new_prefix + string[len(old_prefix) :]  # noqa: E203

    def resolve_string_reference(self, obj: str) -> Any:
        """
        Import an object indicated by ``obj``.

        First, ``obj`` is pre-processed -- prefixes are replaced to allow import:

        - :py:attr:`CUSTOM_DIR_NAMESPACE_PREFIX` is replaced ``{stem}.`` where `stem` is the stem of the custom dir;
        - :py:attr:`CHATSKY_NAMESPACE_PREFIX` is replaced with ``chatsky.``;
        - :py:attr:`EXTERNAL_LIB_NAMESPACE_PREFIX` is removed.

        Next the resulting string is imported:
        If the string is ``a.b.c.d``, the following is tried in order:

        1. ``from a import b; return b.c.d``
        2. ``from a.b import c; return c.d``
        3. ``from a.b.c import d; return d``

        For custom dir imports; parent of the custom dir is appended to ``sys.path`` via :py:meth:`sys_path_append`.

        :return: An imported object.
        :raises ValueError: If ``obj`` does not begin with any of the prefixes (is not :py:meth:`is_resolvable`).
        :raises JSONImportError: If a string could not be imported. Includes exceptions raised on every import attempt.
        """
        # prepare obj string
        if obj.startswith(self.CUSTOM_DIR_NAMESPACE_PREFIX):
            if not self.custom_dir.exists():
                raise JSONImportError(f"Could not find directory {self.custom_dir}")
            obj = self.replace_prefix(obj, self.CUSTOM_DIR_NAMESPACE_PREFIX, self.custom_dir_stem + ".")

        elif obj.startswith(self.CHATSKY_NAMESPACE_PREFIX):
            obj = self.replace_prefix(obj, self.CHATSKY_NAMESPACE_PREFIX, "chatsky.")

        elif obj.startswith(self.EXTERNAL_LIB_NAMESPACE_PREFIX):
            obj = self.replace_prefix(obj, self.EXTERNAL_LIB_NAMESPACE_PREFIX, "")

        else:
            raise ValueError(f"Could not find a namespace prefix: {obj}")

        # import obj
        split = obj.split(".")
        exceptions: List[Exception] = []

        for module_split in range(1, len(split)):
            module_name = ".".join(split[:module_split])
            object_name = split[module_split:]
            try:
                with self.sys_path_append(self.custom_dir_location):
                    module = importlib.import_module(module_name)
                return reduce(getattr, [module, *object_name])
            except Exception as exc:
                exceptions.append(exc)
                logger.debug(f"Exception attempting to import {object_name} from {module_name!r}", exc_info=exc)
        raise JSONImportError(f"Could not import {obj}") from Exception(exceptions)

    def parse_args(self, value: JsonValue) -> Tuple[list, dict]:
        """
        Parse ``value`` into args and kwargs:

        - If ``value`` is a dictionary, it is returned as kwargs;
        - If ``value`` is a list, it is returned as args;
        - If ``value`` is ``None``, both args and kwargs are empty;
        - If ``value`` is anything else, it is returned as the only arg.

        :return: A tuple of args and kwargs.
        """
        args = []
        kwargs = {}
        value = self.replace_resolvable_objects(value)
        if isinstance(value, dict):
            kwargs = value
        elif isinstance(value, list):
            args = value
        elif value is not None:  # none is used when no argument is passed: e.g. `dst.Previous:` does not accept args
            args = [value]

        return args, kwargs

    def replace_resolvable_objects(self, obj: JsonValue) -> Any:
        """
        Replace any resolvable objects inside ``obj`` with their resolved versions and
        initialize any that are the only key of a dictionary.

        This method iterates over every value inside ``obj`` (which is ``JsonValue``).
        Any string that :py:meth:`is_resolvable` is replaced with an object return from
        :py:meth:`resolve_string_reference`.
        This is done only once (i.e. if a string is resolved to another resolvable string,
        that string is not resolved).

        Any dictionaries that contain only one resolvable key are replaced with a result of
        ``resolve_string_reference(key)(*args, **kwargs)`` (the object is initialized)
        where ``args`` and ``kwargs`` is the result of :py:meth:`parse_args`
        on the value of the dictionary.

        :return: A new object with replaced resolvable strings and dictionaries.
        """
        if isinstance(obj, dict):
            keys = obj.keys()
            if len(keys) == 1:
                key = keys.__iter__().__next__()
                if self.is_resolvable(key):
                    args, kwargs = self.parse_args(obj[key])
                    return self.resolve_string_reference(key)(*args, **kwargs)

            return {k: (self.replace_resolvable_objects(v)) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.replace_resolvable_objects(item) for item in obj]
        elif isinstance(obj, str):
            if self.is_resolvable(obj):
                return self.resolve_string_reference(obj)
        return obj

    def import_pipeline_file(self, file: Union[str, Path]) -> dict:
        """
        Import a dictionary from a json/yaml file and replace resolvable objects in it.

        :return: A result of :py:meth:`replace_resolvable_objects` on the dictionary.
        :raises JSONImportError: If a file does not have a correct file extension.
        :raises JSONImportError: If an imported object from file is not a dictionary.
        """
        file = Path(file).absolute()

        with open(file, "r", encoding="utf-8") as fd:
            if file.suffix == ".json":
                pipeline = json.load(fd)
            elif file.suffix in (".yaml", ".yml"):
                if not yaml_available:
                    raise ImportError("`pyyaml` package is missing.\nRun `pip install chatsky[yaml]`.")
                pipeline = yaml.safe_load(fd)
            else:
                raise JSONImportError("File should have a `.json`, `.yaml` or `.yml` extension")
        if not isinstance(pipeline, dict):
            raise JSONImportError("File should contain a dict")

        logger.info(f"Loaded file {file}")
        return self.replace_resolvable_objects(pipeline)


def get_chatsky_objects():
    """
    Return an index of most commonly used ``chatsky`` objects (in the context of pipeline initialization).

    :return: A dictionary where keys are names of the objects (e.g. ``chatsky.core.Message``) and values
        are the objects.
        The items in the dictionary are all the objects from the ``__init__`` files of the following modules:

        - "chatsky.cnd";
        - "chatsky.rsp";
        - "chatsky.dst";
        - "chatsky.proc";
        - "chatsky.core";
        - "chatsky.core.service";
        - "chatsky.slots";
        - "chatsky.context_storages";
        - "chatsky.messengers".
    """
    json_importer = JSONImporter(custom_dir="none")

    def get_objects_from_submodule(submodule_name: str, alias: Optional[str] = None):
        module = json_importer.resolve_string_reference(submodule_name)

        return {
            ".".join([alias or submodule_name, name]): obj
            for name, obj in module.__dict__.items()
            if not name.startswith("_") and not ismodule(obj)
        }

    return {
        k: v
        for module in (
            "chatsky.cnd",
            "chatsky.rsp",
            "chatsky.dst",
            "chatsky.proc",
            "chatsky.core",
            "chatsky.core.service",
            "chatsky.slots",
            "chatsky.context_storages",
            "chatsky.messengers",
            # "chatsky.stats",
            # "chatsky.utils",
        )
        for k, v in get_objects_from_submodule(module).items()
    }



"""
Service Conditions
------------------
Provides service-related conditions
"""

from __future__ import annotations

from chatsky.core.context import Context
from chatsky.core.script_function import BaseCondition

from chatsky.core.service.types import (
    ComponentExecutionState,
)


class ServiceFinished(BaseCondition):
    """
    Check if a :py:class:`~chatsky.core.service.service.Service` was executed successfully.
    """

    path: str
    """The path of the condition pipeline component."""
    wait: bool = False
    """
    Whether to wait for the service to be finished.

    This eliminates possible service states ``NOT_RUN`` and ``RUNNING``.
    """

    def __init__(self, path: str, *, wait: bool = False):
        super().__init__(path=path, wait=wait)

    async def call(self, ctx: Context) -> bool:
        if self.wait:
            await ctx.framework_data.service_states[self.path].finished_event.wait()

        state = ctx.framework_data.service_states[self.path].execution_status

        return ComponentExecutionState[state] == ComponentExecutionState.FINISHED



"""
Service Group
-------------
The Service Group module contains the ServiceGroup class, which is used to represent a group of related services.

This class provides a way to organize and manage multiple services as a single unit,
allowing for easier management and organization of the services within the pipeline.
"""

from __future__ import annotations
import asyncio
import logging
from typing import List, Union, Any, Optional
from typing_extensions import TypeAlias, Annotated

from pydantic import model_validator, Field

from chatsky.core.service.extra import BeforeHandler, AfterHandler
from chatsky.core.script_function import AnyCondition
from chatsky.core.context import Context
from chatsky.core.service.actor import Actor
from chatsky.core.service.component import PipelineComponent
from chatsky.core.service.types import (
    ComponentExecutionState,
    ExtraHandlerType,
    ExtraHandlerConditionFunction,
    ExtraHandlerFunction,
)
from .service import Service, ServiceInitTypes

logger = logging.getLogger(__name__)


class ServiceGroup(PipelineComponent):
    """
    A service group class.
    """

    components: List[
        Union[
            Service,
            ServiceGroup,
        ]
    ]
    """
    A :py:class:`~.ServiceGroup` object, that will be added to the group.
    """
    fully_concurrent: bool = False
    """
    Whether this should run all components inside it concurrently
    (regardless of their `concurrent` attribute.
    This is not recursive (applies only to first level components).
    Default value is False.
    """
    # Repeating inherited fields for better documentation.
    before_handler: BeforeHandler = Field(default_factory=BeforeHandler)
    after_handler: AfterHandler = Field(default_factory=AfterHandler)
    timeout: Optional[float] = None
    concurrent: bool = False
    start_condition: AnyCondition = Field(default=True, validate_default=True)
    name: Optional[str] = None
    path: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def components_validator(cls, data: Any):
        """
        Add support for initializing from a list of :py:class:`~.PipelineComponent`.
        """
        if isinstance(data, list):
            result = {"components": data}
        else:
            result = data

        if isinstance(result, dict):
            if ("components" in result) and (not isinstance(result["components"], list)):
                result["components"] = [result["components"]]
        return result

    async def run_component(self, ctx: Context) -> Optional[ComponentExecutionState]:
        """
        Run components of this service group.

        If :py:attr:`.fully_concurrent` flag is set to True, all :py:attr:`.components` will run concurrently
        (via ``asyncio.gather``).

        Otherwise, all non-concurrent components execute one after another
        while consecutive concurrent components are run concurrently (via ``asyncio.gather``).

        :param ctx: Current dialog context.
        :return: :py:attr:`.ComponentExecutionState.FAILED` if any component failed.
        """
        if self.fully_concurrent:
            await asyncio.gather(*[service(ctx) for service in self.components])
        else:
            current_subgroup = []
            for component in self.components:
                if component.concurrent:
                    current_subgroup.append(component)
                else:
                    await asyncio.gather(*[service(ctx) for service in current_subgroup])
                    await component(ctx)
                    current_subgroup = []
            await asyncio.gather(*[service(ctx) for service in current_subgroup])

        failed = any([service.get_state(ctx) == ComponentExecutionState.FAILED for service in self.components])
        if failed:
            return ComponentExecutionState.FAILED

    def add_extra_handler(
        self,
        extra_handler_type: ExtraHandlerType,
        extra_handler: ExtraHandlerFunction,
        condition: ExtraHandlerConditionFunction = lambda path: False,
    ):
        """
        Add extra handler to this group.

        For every component in the group, ``condition`` is called with the path of that component
        to determine whether to add extra handler to that component.

        :param extra_handler_type: Extra handler type (before or after).
        :param extra_handler: Function to add as an extra handler.
        :type extra_handler: :py:data:`.ExtraHandlerFunction`
        :param condition: Condition function to determine if extra handler should be added to specific subcomponents.
            Defaults to a function returning False.
        :type condition: :py:data:`.ExtraHandlerConditionFunction`
        """
        super().add_extra_handler(extra_handler_type, extra_handler)
        for service in self.components:
            if not condition(service.path):
                continue
            if isinstance(service, ServiceGroup):
                service.add_extra_handler(extra_handler_type, extra_handler, condition)
            else:
                service.add_extra_handler(extra_handler_type, extra_handler)

    @property
    def computed_name(self) -> str:
        """
        "service_group"
        """
        return "service_group"


ServiceGroupInitTypes: TypeAlias = Union[
    ServiceGroup,
    Annotated[List[Union[Actor, ServiceInitTypes, "ServiceGroupInitTypes"]], "list of components"],
    Annotated[Union[Actor, ServiceInitTypes, "ServiceGroupInitTypes"], "single component of the group"],
    Annotated[dict, "dict following the ServiceGroup data model"],
]
"""Types that :py:class:`~.ServiceGroup` can be validated from."""



"""
Utils
-----
The Utils module contains functions used to provide names to nameless pipeline components inside of a group.
"""

import collections
from typing import List

from chatsky.core import Context
from chatsky.core.context import ServiceState
from .service.component import PipelineComponent
from .service.group import ServiceGroup


def rename_component_incrementing(index: int, computed_names: List[str]) -> str:
    """
    Function for generating new name for a pipeline component,
    that has similar name with other components in the same group.

    The name is generated according to these rules:

    1. Base name is :py:attr:`.PipelineComponent.computed_name`;
    2. If there are multiple components with the same name, ``#[NUMBER]`` is added to the resulting name,
       where ``NUMBER`` is the number of components with the same name in current service group.

    :param index: Index of the component to be renamed.
    :param computed_names: A list of component names (or computed names).
    :return: Name for the component.
    """
    base_name = computed_names[index]
    if computed_names.count(base_name) == 1:
        return base_name
    else:
        return f"{base_name}#{computed_names[:index].count(base_name)}"


def finalize_service_group(service_group: ServiceGroup, path: str = "") -> None:
    """
    Function that iterates through a service group (and all its subgroups),
    finalizing component's names and paths in it.
    Components are renamed only if user didn't set a name for them. Their paths are also generated here.

    :param service_group: Service group to resolve name collisions in.
    :param path:
        A prefix for component paths -- path of `component` is equal to `{path}.{component.name}`.
        Defaults to "".
    :raises ValueError: If multiple components have the same name assigned to them.
    """
    computed_names = [
        component.name if component.name is not None else component.computed_name
        for component in service_group.components
    ]
    for idx, component in enumerate(service_group.components):
        if component.name is None:
            component.name = rename_component_incrementing(idx, computed_names)
        component.path = f"{path}.{component.name}"

    new_names_counter = collections.Counter([component.name for component in service_group.components])
    for k, v in new_names_counter.items():
        if v != 1:
            raise ValueError("Found multiple components with the same name: {path}.{k}")

    for component in service_group.components:
        if isinstance(component, ServiceGroup):
            finalize_service_group(component, f"{path}.{component.name}")


def initialize_service_states(ctx: Context, service: PipelineComponent) -> None:
    """
    Reset :py:class:`.ServiceState` of `service`.

    Called at the beginning of every turn for the pipeline service group.
    """
    ctx.framework_data.service_states[service.path] = ServiceState()
    if isinstance(service, ServiceGroup):
        for component in service.components:
            initialize_service_states(ctx, component)



"""
Slot Responses
--------------
Slot-related responses.
"""

from typing import Union, Literal
import logging

from chatsky.core import Context, Message, BaseResponse
from chatsky.core.script_function import AnyResponse
from chatsky.core.message import MessageInitTypes


logger = logging.getLogger(__name__)


class FilledTemplate(BaseResponse):
    """
    Fill template with slot values.
    The `text` attribute of the template message should be a format-string:
    e.g. "Your username is {profile.username}".

    For the example above, if ``profile.username`` slot has value "admin",
    it would return a copy of the message with the following text:
    "Your username is admin".
    """

    template: AnyResponse
    """A response to use as a template."""
    on_exception: Literal["keep_template", "return_none"] = "return_none"
    """
    What to do if template filling fails.

    - "keep_template": :py:attr:`template` is returned, unfilled.
    - "return_none": an empty message is returned.
    """

    def __init__(
        self,
        template: Union[MessageInitTypes, BaseResponse],
        on_exception: Literal["keep_template", "return_none"] = "return_none",
    ):
        super().__init__(template=template, on_exception=on_exception)

    async def call(self, ctx: Context) -> MessageInitTypes:
        result = await self.template(ctx)

        if result.text is not None:
            filled = ctx.framework_data.slot_manager.fill_template(result.text)
            if isinstance(filled, str):
                result.text = filled
                return result
            else:
                if self.on_exception == "return_none":
                    return Message()
                else:
                    return result
        else:
            logger.warning(f"`template` of `FilledTemplate` returned `Message` without `text`: {result}.")
            return result



"""
Script Function
---------------
This module provides base classes for functions used in :py:class:`~chatsky.core.script.Script` instances.

These functions allow dynamic script configuration and are essential to the scripting process.
"""

from __future__ import annotations

from typing import Union, Tuple, ClassVar, Optional
from typing_extensions import Annotated
from abc import abstractmethod, ABC
import logging

from pydantic import BaseModel, model_validator, Field

from chatsky.utils.devel import wrap_sync_function_in_async
from chatsky.core.context import Context
from chatsky.core.message import Message, MessageInitTypes
from chatsky.core.node_label import NodeLabel, NodeLabelInitTypes, AbsoluteNodeLabel


logger = logging.getLogger(__name__)


class BaseScriptFunc(BaseModel, ABC, frozen=True):  # generic doesn't work well with sphinx autosummary
    """
    Base class for any script function.

    Defines :py:meth:`wrapped_call` that wraps :py:meth:`call` and handles exceptions and types conversions.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]]
    """Return type of the script function."""

    @abstractmethod
    async def call(self, ctx: Context):
        """Implement this to create a custom function."""
        raise NotImplementedError()

    async def wrapped_call(self, ctx: Context, *, info: str = ""):
        """
        Exception-safe wrapper for :py:meth:`__call__`.

        :return: An instance of :py:attr:`return_type` if possible.
            Otherwise, an ``Exception`` instance detailing what went wrong.
        """
        try:
            result = await self(ctx)
            logger.debug(f"Function {self.__class__.__name__} returned {result!r}. {info}")
            return result
        except Exception as exc:
            logger.error(f"An exception occurred in {self.__class__.__name__}. {info}", exc_info=exc)
            return exc

    async def __call__(self, ctx: Context):
        """
        Handle :py:meth:`call`:

        - Call it (regardless of whether it is async);
        - Cast returned value to :py:attr:`return_type`.

        :return: An instance of :py:attr:`return_type`.
        :raises TypeError: If :py:meth:`call` returned value of incorrect type.
        """
        result = await wrap_sync_function_in_async(self.call, ctx)
        if not isinstance(self.return_type, tuple) and issubclass(self.return_type, BaseModel):
            result = self.return_type.model_validate(result, context={"ctx": ctx}).model_copy(deep=True)
        if not isinstance(result, self.return_type):
            raise TypeError(
                f"Function `call` of {self.__class__.__name__} should return {self.return_type!r}. "
                f"Got instead: {result!r}"
            )
        return result


class ConstScriptFunc(BaseScriptFunc):
    """
    Base class for script functions that return a constant value.
    """

    root: None
    """Value to return."""

    async def call(self, ctx: Context):
        return self.root

    @model_validator(mode="before")
    @classmethod
    def validate_value(cls, data):
        """Allow instantiating this class from its root value."""
        return {"root": data}


class BaseCondition(BaseScriptFunc, ABC):
    """
    Base class for condition functions.

    These are used in :py:attr:`chatsky.core.transition.Transition.cnd`.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]] = bool

    @abstractmethod
    async def call(self, ctx: Context) -> bool:
        raise NotImplementedError

    async def wrapped_call(self, ctx: Context, *, info: str = "") -> Union[bool, Exception]:
        return await super().wrapped_call(ctx, info=info)

    async def __call__(self, ctx: Context) -> bool:
        return await super().__call__(ctx)

    async def is_true(self, ctx: Context, *, info: str = "") -> bool:
        """Same as :py:meth:`wrapped_call` but instead of exceptions return ``False``."""
        result = await self.wrapped_call(ctx, info=info)
        if isinstance(result, Exception):
            return False
        return result


class ConstCondition(ConstScriptFunc, BaseCondition):
    root: bool


AnyCondition = Annotated[Union[ConstCondition, BaseCondition], Field(union_mode="left_to_right")]
"""
A type annotation that allows accepting both :py:class:`ConstCondition` and :py:class:`BaseCondition`
while validating :py:class:`ConstCondition` if possible.
"""


class BaseResponse(BaseScriptFunc, ABC):
    """
    Base class for response functions.

    These are used in :py:attr:`chatsky.core.script.Node.response`.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]] = Message

    @abstractmethod
    async def call(self, ctx: Context) -> MessageInitTypes:
        raise NotImplementedError

    async def wrapped_call(self, ctx: Context, *, info: str = "") -> Union[Message, Exception]:
        return await super().wrapped_call(ctx, info=info)

    async def __call__(self, ctx: Context) -> Message:
        return await super().__call__(ctx)


class ConstResponse(ConstScriptFunc, BaseResponse):
    root: Message


AnyResponse = Annotated[Union[ConstResponse, BaseResponse], Field(union_mode="left_to_right")]
"""
A type annotation that allows accepting both :py:class:`ConstResponse` and :py:class:`BaseResponse`
while validating :py:class:`ConstResponse` if possible.
"""


class BaseDestination(BaseScriptFunc, ABC):
    """
    Base class for destination functions.

    These are used in :py:attr:`chatsky.core.transition.Transition.dst`.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]] = AbsoluteNodeLabel

    @abstractmethod
    async def call(self, ctx: Context) -> NodeLabelInitTypes:
        raise NotImplementedError

    async def wrapped_call(self, ctx: Context, *, info: str = "") -> Union[AbsoluteNodeLabel, Exception]:
        return await super().wrapped_call(ctx, info=info)

    async def __call__(self, ctx: Context) -> AbsoluteNodeLabel:
        return await super().__call__(ctx)


class ConstDestination(ConstScriptFunc, BaseDestination):
    root: NodeLabel


AnyDestination = Annotated[Union[ConstDestination, BaseDestination], Field(union_mode="left_to_right")]
"""
A type annotation that allows accepting both :py:class:`ConstDestination` and :py:class:`BaseDestination`
while validating :py:class:`ConstDestination` if possible.
"""


class BaseProcessing(BaseScriptFunc, ABC):
    """
    Base class for processing functions.

    These are used in :py:attr:`chatsky.core.script.Node.pre_transition`
    and :py:attr:`chatsky.core.script.Node.pre_response`.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]] = type(None)

    @abstractmethod
    async def call(self, ctx: Context) -> None:
        raise NotImplementedError

    async def wrapped_call(self, ctx: Context, *, info: str = "") -> Union[None, Exception]:
        return await super().wrapped_call(ctx, info=info)

    async def __call__(self, ctx: Context) -> None:
        return await super().__call__(ctx)


class BasePriority(BaseScriptFunc, ABC):
    """
    Base class for priority functions.

    These are used in :py:attr:`chatsky.core.transition.Transition.priority`.

    Has several possible return types:

    - ``float``: Transition successful with the corresponding priority;
    - ``True`` or ``None``: Transition successful with the :py:attr:`~chatsky.core.pipeline.Pipeline.default_priority`;
    - ``False``: Transition unsuccessful.
    """

    return_type: ClassVar[Union[type, Tuple[type, ...]]] = (float, type(None), bool)

    @abstractmethod
    async def call(self, ctx: Context) -> Union[float, bool, None]:
        raise NotImplementedError

    async def wrapped_call(self, ctx: Context, *, info: str = "") -> Union[float, bool, None, Exception]:
        return await super().wrapped_call(ctx, info=info)

    async def __call__(self, ctx: Context) -> Union[float, bool, None]:
        return await super().__call__(ctx)


class ConstPriority(ConstScriptFunc, BasePriority):
    root: Optional[float]


AnyPriority = Annotated[Union[ConstPriority, BasePriority], Field(union_mode="left_to_right")]
"""
A type annotation that allows accepting both :py:class:`ConstPriority` and :py:class:`BasePriority`
while validating :py:class:`ConstPriority` if possible.
"""



"""
Slot Processing
---------------
This module provides wrappers for :py:class:`~chatsky.slots.slots.SlotManager`'s API as :py:class:`.BaseProcessing`
subclasses.
"""

import asyncio
import logging
from typing import List

from chatsky.slots.slots import SlotName
from chatsky.core import Context, BaseProcessing
from chatsky.responses.slots import FilledTemplate

logger = logging.getLogger(__name__)


class Extract(BaseProcessing):
    """
    Extract slots listed slots.
    This will override all slots even if they are already extracted.
    """

    slots: List[SlotName]
    """A list of slot names to extract."""
    success_only: bool = True
    """If set, only successfully extracted values will be stored in the slot storage."""

    def __init__(self, *slots: SlotName, success_only: bool = True):
        super().__init__(slots=slots, success_only=success_only)

    async def call(self, ctx: Context):
        manager = ctx.framework_data.slot_manager
        results = await asyncio.gather(
            *(manager.extract_slot(slot, ctx, self.success_only) for slot in self.slots), return_exceptions=True
        )

        for result in results:
            if isinstance(result, Exception):
                logger.exception("An exception occurred during slot extraction.", exc_info=result)


class Unset(BaseProcessing):
    """
    Mark specified slots as not extracted and clear extracted values.
    """

    slots: List[SlotName]
    """A list of slot names to extract."""

    def __init__(self, *slots: SlotName):
        super().__init__(slots=slots)

    async def call(self, ctx: Context):
        manager = ctx.framework_data.slot_manager
        for slot in self.slots:
            try:
                manager.unset_slot(slot)
            except Exception as exc:
                logger.exception("An exception occurred during slot resetting.", exc_info=exc)


class UnsetAll(BaseProcessing):
    """
    Mark all slots as not extracted and clear all extracted values.
    """

    async def call(self, ctx: Context):
        manager = ctx.framework_data.slot_manager
        manager.unset_all_slots()


class FillTemplate(BaseProcessing):
    """
    Fill the response template in the current node.

    Response message of the current node should be a format-string: e.g. ``"Your username is {profile.username}"``.
    """

    async def call(self, ctx: Context):
        response = ctx.current_node.response

        if response is None:
            return

        ctx.current_node.response = FilledTemplate(template=response)



"""
Transition
----------
This module defines a transition class that is used to
specify conditions and destinations for transitions to nodes.
"""

from __future__ import annotations

from typing import Union, List, TYPE_CHECKING, Optional, Tuple
import logging
import asyncio

from pydantic import BaseModel, Field

from chatsky.core.script_function import AnyCondition, AnyDestination, AnyPriority
from chatsky.core.script_function import BaseCondition, BaseDestination, BasePriority
from chatsky.core.node_label import AbsoluteNodeLabel, NodeLabelInitTypes

if TYPE_CHECKING:
    from chatsky.core.context import Context


logger = logging.getLogger(__name__)


class Transition(BaseModel):
    """
    A basic class for a transition to a node.
    """

    cnd: AnyCondition = Field(default=True, validate_default=True)
    """A condition that determines if transition is allowed to happen."""
    dst: AnyDestination
    """Destination node of the transition."""
    priority: AnyPriority = Field(default=None, validate_default=True)
    """Priority of the transition. Higher priority transitions are resolved first."""

    def __init__(
        self,
        *,
        cnd: Union[bool, BaseCondition] = True,
        dst: Union[NodeLabelInitTypes, BaseDestination],
        priority: Union[Optional[float], BasePriority] = None,
    ):
        super().__init__(cnd=cnd, dst=dst, priority=priority)


async def get_next_label(
    ctx: Context, transitions: List[Transition], default_priority: float
) -> Optional[AbsoluteNodeLabel]:
    """
    Determine the next node based on ``transitions`` and ``ctx``.

    The process is as follows:

    1. Condition result is calculated for every transition.
    2. Transitions are filtered by the calculated condition.
    3. Priority result is calculated for every transition that is left.
       ``default_priority`` is used for priorities that return ``True`` or ``None``
       as per :py:class:`.BasePriority`.
       Those that return ``False`` are filtered out.
    4. Destination result is calculated for every transition that is left.
    5. The highest priority transition is chosen.
       If there are multiple transition of the higher priority,
       choose the first one of that priority in the ``transitions`` list.
       Order of ``transitions`` is as follows:
       ``node transitions, local transitions, global transitions``.

    If at any point any :py:class:`.BaseCondition`, :py:class:`.BaseDestination` or :py:class:`.BasePriority`
    produces an exception, the corresponding transition is filtered out.

    :return: Label of the next node or ``None`` if no transition is left by the end of the process.
    """
    filtered_transitions: List[Transition] = transitions.copy()
    condition_results = await asyncio.gather(*[transition.cnd.wrapped_call(ctx) for transition in filtered_transitions])

    filtered_transitions = [
        transition for transition, condition in zip(filtered_transitions, condition_results) if condition is True
    ]

    priority_results = await asyncio.gather(
        *[transition.priority.wrapped_call(ctx) for transition in filtered_transitions]
    )

    transitions_with_priorities: List[Tuple[Transition, float]] = [
        (transition, (priority_result if isinstance(priority_result, float) else default_priority))
        for transition, priority_result in zip(filtered_transitions, priority_results)
        if (priority_result is True or priority_result is None or isinstance(priority_result, float))
    ]
    logger.debug(f"Possible transitions: {transitions_with_priorities!r}")

    transitions_with_priorities = sorted(transitions_with_priorities, key=lambda x: x[1], reverse=True)

    destination_results = await asyncio.gather(
        *[transition.dst.wrapped_call(ctx) for transition, _ in transitions_with_priorities]
    )

    for destination in destination_results:
        if isinstance(destination, AbsoluteNodeLabel):
            return destination
    return None



"""
Standard Conditions
-------------------
This module provides basic conditions.

- :py:class:`.Any`, :py:class:`.All` and :py:class:`.Negation` are meta-conditions.
- :py:class:`.HasText`, :py:class:`.Regexp`, :py:class:`.HasCallbackQuery` are last-request-based conditions.
- :py:class:`.CheckLastLabels` is a label-based condition.
"""

import asyncio
from typing import Pattern, Union, List, Optional
import logging
import re
from functools import cached_property

from pydantic import Field, computed_field

from chatsky.core import BaseCondition, Context
from chatsky.core.message import Message, MessageInitTypes, CallbackQuery
from chatsky.core.node_label import AbsoluteNodeLabel, AbsoluteNodeLabelInitTypes

logger = logging.getLogger(__name__)


class ExactMatch(BaseCondition):
    """
    Check if :py:attr:`~.Context.last_request` matches :py:attr:`.match`.

    If :py:attr:`.skip_none`, will not compare ``None`` fields of :py:attr:`.match`.
    """

    match: Message
    """
    Message to compare last request with.

    Is initialized according to :py:data:`~.MessageInitTypes`.
    """
    skip_none: bool = True
    """
    Whether fields set to ``None`` in :py:attr:`.match` should not be compared.
    """

    def __init__(self, match: MessageInitTypes, *, skip_none=True):
        super().__init__(match=match, skip_none=skip_none)

    async def call(self, ctx: Context) -> bool:
        request = ctx.last_request
        for field in self.match.model_fields:
            match_value = self.match.__getattribute__(field)
            if self.skip_none and match_value is None:
                continue
            if field in request.model_fields.keys():
                if request.__getattribute__(field) != self.match.__getattribute__(field):
                    return False
            else:
                return False
        return True


class HasText(BaseCondition):
    """
    Check if the :py:attr:`~.Message.text` attribute of :py:attr:`~.Context.last_request`
    contains :py:attr:`.text`.
    """

    text: str
    """
    Text to search for in the last request.
    """

    def __init__(self, text: str):
        super().__init__(text=text)

    async def call(self, ctx: Context) -> bool:
        request = ctx.last_request
        if request.text is None:
            return False
        return self.text in request.text


class Regexp(BaseCondition):
    """
    Check if the :py:attr:`~.Message.text` attribute of :py:attr:`~.Context.last_request`
    contains :py:attr:`.pattern`.
    """

    pattern: Union[str, Pattern]
    """
    The `RegExp` pattern to search for in the last request.
    """
    flags: Union[int, re.RegexFlag] = 0
    """
    Flags to pass to ``re.compile``.
    """

    def __init__(self, pattern: Union[str, Pattern], *, flags: Union[int, re.RegexFlag] = 0):
        super().__init__(pattern=pattern, flags=flags)

    @computed_field
    @cached_property
    def re_object(self) -> Pattern:
        """Compiled pattern."""
        return re.compile(self.pattern, self.flags)

    async def call(self, ctx: Context) -> bool:
        request = ctx.last_request
        if request.text is None:
            return False
        return bool(self.re_object.search(request.text))


class Any(BaseCondition):
    """
    Check if any condition from the :py:attr:`.conditions` list is True.
    """

    conditions: List[BaseCondition]
    """
    List of conditions.
    """

    def __init__(self, *conditions: BaseCondition):
        super().__init__(conditions=list(conditions))

    async def call(self, ctx: Context) -> bool:
        return any(await asyncio.gather(*(cnd.is_true(ctx) for cnd in self.conditions)))


class All(BaseCondition):
    """
    Check if all conditions from the :py:attr:`.conditions` list is True.
    """

    conditions: List[BaseCondition]
    """
    List of conditions.
    """

    def __init__(self, *conditions: BaseCondition):
        super().__init__(conditions=list(conditions))

    async def call(self, ctx: Context) -> bool:
        return all(await asyncio.gather(*(cnd.is_true(ctx) for cnd in self.conditions)))


class Negation(BaseCondition):
    """
    Return the negation of the result of :py:attr:`~.Negation.condition`.
    """

    condition: BaseCondition
    """
    Condition to negate.
    """

    def __init__(self, condition: BaseCondition):
        super().__init__(condition=condition)

    async def call(self, ctx: Context) -> bool:
        return not await self.condition.is_true(ctx)


Not = Negation
"""
:py:class:`.Not` is an alias for :py:class:`.Negation`.
"""


class CheckLastLabels(BaseCondition):
    """
    Check if any label in the last :py:attr:`.last_n_indices` of :py:attr:`.Context.labels` is in
    :py:attr:`.labels` or if its :py:attr:`~.AbsoluteNodeLabel.flow_name` is in :py:attr:`.flow_labels`.
    """

    flow_labels: List[str] = Field(default_factory=list)
    """
    List of flow names to find in the last labels.
    """
    labels: List[AbsoluteNodeLabel] = Field(default_factory=list)
    """
    List of labels to find in the last labels.

    Is initialized according to :py:data:`~.AbsoluteNodeLabelInitTypes`.
    """
    last_n_indices: int = Field(default=1, ge=1)
    """
    Number of labels to check.
    """

    def __init__(
        self,
        *,
        flow_labels: Optional[List[str]] = None,
        labels: Optional[List[AbsoluteNodeLabelInitTypes]] = None,
        last_n_indices: int = 1
    ):
        if flow_labels is None:
            flow_labels = []
        if labels is None:
            labels = []
        super().__init__(flow_labels=flow_labels, labels=labels, last_n_indices=last_n_indices)

    async def call(self, ctx: Context) -> bool:
        labels = list(ctx.labels.values())[-self.last_n_indices :]  # noqa: E203
        for label in labels:
            if label.flow_name in self.flow_labels or label in self.labels:
                return True
        return False


class HasCallbackQuery(BaseCondition):
    """
    Check if :py:attr:`~.Context.last_request` contains a :py:class:`.CallbackQuery` attachment
    with :py:attr:`.CallbackQuery.query_string` matching :py:attr:`.HasCallbackQuery.query_string`.
    """

    query_string: str
    """
    Query string to find in last request's attachments.
    """

    def __init__(self, query_string: str):
        super().__init__(query_string=query_string)

    async def call(self, ctx: Context) -> bool:
        last_request = ctx.last_request
        if last_request.attachments is None:
            return False
        for attachment in last_request.attachments:
            if isinstance(attachment, CallbackQuery):
                if attachment.query_string == self.query_string:
                    return True
        return False



# flake8: noqa: F401

from chatsky.core.service.types import ExtraHandlerRuntimeInfo, ComponentExecutionState
from chatsky.core import Context, Script
from chatsky.core.script import Node
from chatsky.core.pipeline import Pipeline
from chatsky.slots.slots import SlotManager
from chatsky.core.context import FrameworkData, ServiceState
from chatsky.core.service import PipelineComponent

PipelineComponent.model_rebuild()
Pipeline.model_rebuild()
Script.model_rebuild()
Context.model_rebuild()
ExtraHandlerRuntimeInfo.model_rebuild()
FrameworkData.model_rebuild()
ServiceState.model_rebuild()



"""
Slot Conditions
---------------------------
Provides slot-related conditions.
"""

from __future__ import annotations
from typing import Literal, List

from chatsky.core import Context, BaseCondition
from chatsky.slots.slots import SlotName


class SlotsExtracted(BaseCondition):
    """
    Check if :py:attr:`.slots` are extracted.

    :param mode: Whether to check if all slots are extracted or any slot is extracted.
    """

    slots: List[SlotName]
    """
    Names of the slots that need to be checked.
    """
    mode: Literal["any", "all"] = "all"
    """
    Whether to check if all slots are extracted or any slot is extracted.
    """

    def __init__(self, *slots: SlotName, mode: Literal["any", "all"] = "all"):
        super().__init__(slots=slots, mode=mode)

    async def call(self, ctx: Context) -> bool:
        manager = ctx.framework_data.slot_manager
        if self.mode == "all":
            return all(manager.is_slot_extracted(slot) for slot in self.slots)
        elif self.mode == "any":
            return any(manager.is_slot_extracted(slot) for slot in self.slots)



"""
Context
-------
Context is a data structure that is used to store information about the current state of a conversation.

It is used to keep track of the user's input, the current stage of the conversation, and any other
information that is relevant to the current context of a dialog.

The Context data structure provides several key features to make working with data easier.
Developers can use the context to store any information that is relevant to the current conversation,
such as user data, session data, conversation history, e.t.c.
This allows developers to easily access and use this data throughout the conversation flow.

Another important feature of the context is data serialization.
The context can be easily serialized to a format that can be stored or transmitted, such as JSON.
This allows developers to save the context data and resume the conversation later.
"""

from __future__ import annotations
import logging
import asyncio
from uuid import UUID, uuid4
from typing import Any, Optional, Union, Dict, TYPE_CHECKING

from pydantic import BaseModel, Field

from chatsky.core.message import Message, MessageInitTypes
from chatsky.slots.slots import SlotManager
from chatsky.core.node_label import AbsoluteNodeLabel, AbsoluteNodeLabelInitTypes

if TYPE_CHECKING:
    from chatsky.core.service import ComponentExecutionState
    from chatsky.core.script import Node
    from chatsky.core.pipeline import Pipeline

logger = logging.getLogger(__name__)


def get_last_index(dictionary: dict) -> int:
    """
    Obtain the last index from the `dictionary`.

    :param dictionary: Dictionary with unsorted keys.
    :return: Last index from the `dictionary`.
    :raises ValueError: If the dictionary is empty.
    """
    if len(dictionary) == 0:
        raise ValueError("Dictionary is empty.")
    indices = list(dictionary)
    return max(indices)


class ContextError(Exception):
    """Raised when context methods are not used correctly."""


class ServiceState(BaseModel, arbitrary_types_allowed=True):
    execution_status: ComponentExecutionState = Field(default="NOT_RUN")
    """
    :py:class:`.ComponentExecutionState` of this pipeline service.
    Cleared at the end of every turn.
    """
    finished_event: asyncio.Event = Field(default_factory=asyncio.Event)
    """
    Asyncio `Event` which can be awaited until this service finishes.
    Cleared at the end of every turn.
    """


class FrameworkData(BaseModel, arbitrary_types_allowed=True):
    """
    Framework uses this to store data related to any of its modules.
    """

    service_states: Dict[str, ServiceState] = Field(default_factory=dict, exclude=True)
    """
    Dictionary containing :py:class:`.ServiceState` of all the pipeline components.
    Cleared at the end of every turn.
    """
    current_node: Optional[Node] = Field(default=None, exclude=True)
    """
    A copy of the current node provided by :py:meth:`~chatsky.core.script.Script.get_inherited_node`.
    This node can be safely modified by Processing functions to alter current node fields.
    """
    pipeline: Optional[Pipeline] = Field(default=None, exclude=True)
    """
    Instance of the pipeline that manages this context.
    Can be used to obtain run configuration such as script or fallback label.
    """
    stats: Dict[str, Any] = Field(default_factory=dict)
    "Enables complex stats collection across multiple turns."
    slot_manager: SlotManager = Field(default_factory=SlotManager)
    "Stores extracted slots."


class Context(BaseModel):
    """
    A structure that is used to store data about the context of a dialog.
    """

    id: Union[UUID, int, str] = Field(default_factory=uuid4)
    """
    ``id`` is the unique context identifier. By default, randomly generated using ``uuid4``.
    ``id`` can be used to trace the user behavior, e.g while collecting the statistical data.
    """
    labels: Dict[int, AbsoluteNodeLabel] = Field(default_factory=dict)
    """
    ``labels`` stores the history of labels for all passed nodes.

        - key - ``id`` of the turn.
        - value - ``label`` of this turn.

    Start label is stored at key ``0``.
    IDs go up by ``1`` after that.
    """
    requests: Dict[int, Message] = Field(default_factory=dict)
    """
    ``requests`` stores the history of all requests received by the pipeline.

        - key - ``id`` of the turn.
        - value - ``request`` of this turn.

    First request is stored at key ``1``.
    IDs go up by ``1`` after that.
    """
    responses: Dict[int, Message] = Field(default_factory=dict)
    """
    ``responses`` stores the history of all responses produced by the pipeline.

        - key - ``id`` of the turn.
        - value - ``response`` of this turn.

    First response is stored at key ``1``.
    IDs go up by ``1`` after that.
    """
    misc: Dict[str, Any] = Field(default_factory=dict)
    """
    ``misc`` stores any custom data. The framework doesn't use this dictionary,
    so storage of any data won't reflect on the work of the internal Chatsky functions.

        - key - Arbitrary data name.
        - value - Arbitrary data.
    """
    framework_data: FrameworkData = Field(default_factory=FrameworkData)
    """
    This attribute is used for storing custom data required for pipeline execution.
    It is meant to be used by the framework only. Accessing it may result in pipeline breakage.
    """

    @classmethod
    def init(cls, start_label: AbsoluteNodeLabelInitTypes, id: Optional[Union[UUID, int, str]] = None):
        """Initialize new context from ``start_label`` and, optionally, context ``id``."""
        init_kwargs = {
            "labels": {0: AbsoluteNodeLabel.model_validate(start_label)},
        }
        if id is None:
            return cls(**init_kwargs)
        else:
            return cls(**init_kwargs, id=id)

    def add_request(self, request: MessageInitTypes):
        """
        Add a new ``request`` to the context.
        """
        request_message = Message.model_validate(request)
        if len(self.requests) == 0:
            self.requests[1] = request_message
        else:
            last_index = get_last_index(self.requests)
            self.requests[last_index + 1] = request_message

    def add_response(self, response: MessageInitTypes):
        """
        Add a new ``response`` to the context.
        """
        response_message = Message.model_validate(response)
        if len(self.responses) == 0:
            self.responses[1] = response_message
        else:
            last_index = get_last_index(self.responses)
            self.responses[last_index + 1] = response_message

    def add_label(self, label: AbsoluteNodeLabelInitTypes):
        """
        Add a new :py:class:`~.AbsoluteNodeLabel` to the context.

        :raises ContextError: If :py:attr:`labels` is empty.
        """
        label = AbsoluteNodeLabel.model_validate(label)
        if len(self.labels) == 0:
            raise ContextError("Labels are empty. Use `Context.init` to initialize context with labels.")
        last_index = get_last_index(self.labels)
        self.labels[last_index + 1] = label

    @property
    def last_label(self) -> AbsoluteNodeLabel:
        """
        Return the last :py:class:`~.AbsoluteNodeLabel` of
        the :py:class:`~.Context`.

        :raises ContextError: If :py:attr:`labels` is empty.
        """
        if len(self.labels) == 0:
            raise ContextError("Labels are empty. Use `Context.init` to initialize context with labels.")
        last_index = get_last_index(self.labels)
        return self.labels[last_index]

    @property
    def last_response(self) -> Optional[Message]:
        """
        Return the last response of the current :py:class:`~.Context`.
        Return ``None`` if no responses have been added yet.
        """
        if len(self.responses) == 0:
            return None
        last_index = get_last_index(self.responses)
        response = self.responses[last_index]
        return response

    @property
    def last_request(self) -> Message:
        """
        Return the last request of the current :py:class:`~.Context`.

        :raises ContextError: If :py:attr:`responses` is empty.
        """
        if len(self.requests) == 0:
            raise ContextError("No requests have been added.")
        last_index = get_last_index(self.requests)
        return self.requests[last_index]

    @property
    def pipeline(self) -> Pipeline:
        """Return :py:attr:`.FrameworkData.pipeline`."""
        pipeline = self.framework_data.pipeline
        if pipeline is None:
            raise ContextError("Pipeline is not set.")
        return pipeline

    @property
    def current_node(self) -> Node:
        """Return :py:attr:`.FrameworkData.current_node`."""
        node = self.framework_data.current_node
        if node is None:
            raise ContextError("Current node is not set.")
        return node
